"""Test for `maas.client.viscera.static_routes`."""

import random

from testtools.matchers import Equals

from ..static_routes import StaticRoute, StaticRoutes

from ..testing import bind
from ...testing import make_string_without_spaces, TestCase


def make_origin():
    """
    Create a new origin with StaticRoutes and StaticRoute. The former
    refers to the latter via the origin, hence why it must be bound.
    """
    return bind(StaticRoutes, StaticRoute)


class TestStaticRoutes(TestCase):
    def test__static_routes_create(self):
        StaticRoutes = make_origin().StaticRoutes
        destination = random.randint(0, 100)
        source = random.randint(0, 100)
        gateway_ip = make_string_without_spaces()
        metric = make_string_without_spaces()
        StaticRoutes._handler.create.return_value = {
            "id": 1,
            "destination": destination,
            "source": source,
            "gateway_ip": gateway_ip,
            "metric": metric,
        }
        StaticRoutes.create(
            destination=destination, source=source, gateway_ip=gateway_ip, metric=metric
        )
        StaticRoutes._handler.create.assert_called_once_with(
            destination=destination, source=source, gateway_ip=gateway_ip, metric=metric
        )

    def test__static_routes_read(self):
        """StaticRoutes.read() returns a list of StaticRoutes."""
        StaticRoutes = make_origin().StaticRoutes
        static_routes = [
            {
                "id": random.randint(0, 100),
                "destination": random.randint(0, 100),
                "source": random.randint(0, 100),
                "gateway_ip": make_string_without_spaces(),
                "metric": make_string_without_spaces(),
            }
            for _ in range(3)
        ]
        StaticRoutes._handler.read.return_value = static_routes
        static_routes = StaticRoutes.read()
        self.assertThat(len(static_routes), Equals(3))


class TestStaticRoute(TestCase):
    def test__static_route_read(self):
        StaticRoute = make_origin().StaticRoute
        static_route = {
            "id": random.randint(0, 100),
            "destination": random.randint(0, 100),
            "source": random.randint(0, 100),
            "gateway_ip": make_string_without_spaces(),
            "metric": make_string_without_spaces(),
        }
        StaticRoute._handler.read.return_value = static_route
        self.assertThat(
            StaticRoute.read(id=static_route["id"]), Equals(StaticRoute(static_route))
        )
        StaticRoute._handler.read.assert_called_once_with(id=static_route["id"])

    def test__static_route_delete(self):
        StaticRoute = make_origin().StaticRoute
        static_route_id = random.randint(1, 100)
        static_route = StaticRoute(
            {
                "id": static_route_id,
                "destination": random.randint(0, 100),
                "source": random.randint(0, 100),
                "gateway_ip": make_string_without_spaces(),
                "metric": make_string_without_spaces(),
            }
        )
        static_route.delete()
        StaticRoute._handler.delete.assert_called_once_with(id=static_route_id)



"""Commands for nodes."""

__all__ = ["register"]

from . import CommandError, OriginPagedTableCommand, tables
from ..enum import NodeType


class cmd_nodes(OriginPagedTableCommand):
    """List nodes."""

    def __init__(self, parser):
        super(cmd_nodes, self).__init__(parser)
        parser.add_argument("hostname", nargs="*", help=("Hostname of the node."))

    def execute(self, origin, options, target):
        hostnames = None
        if options.hostname:
            hostnames = options.hostname
        table = tables.NodesTable()
        return table.render(target, origin.Nodes.read(hostnames=hostnames))


class cmd_node(OriginPagedTableCommand):
    """Details of a node."""

    def __init__(self, parser):
        super(cmd_node, self).__init__(parser)
        parser.add_argument("hostname", nargs=1, help=("Hostname of the node."))

    def execute(self, origin, options, target):
        nodes = origin.Nodes.read(hostnames=options.hostname)
        if len(nodes) == 0:
            raise CommandError("Unable to find node %s." % options.hostname[0])
        node = nodes[0]
        if node.node_type == NodeType.MACHINE:
            table = tables.MachineDetail(with_type=True)
            node = node.as_machine()
        elif node.node_type == NodeType.DEVICE:
            table = tables.DeviceDetail(with_type=True)
            node = node.as_device()
        elif node.node_type == NodeType.RACK_CONTROLLER:
            table = tables.ControllerDetail()
            node = node.as_rack_controller()
        elif node.node_type == NodeType.REGION_CONTROLLER:
            table = tables.ControllerDetail()
            node = node.as_region_controller()
        elif node.node_type == NodeType.REGION_AND_RACK_CONTROLLER:
            table = tables.ControllerDetail()
            node = node.as_rack_controller()
        return table.render(target, node)


def register(parser):
    """Register commands with the given parser."""
    cmd_nodes.register(parser)
    cmd_node.register(parser)



"""Test for `maas.client.viscera.boot_sources`."""

import random

from testtools.matchers import Equals, MatchesStructure

from .. import boot_sources
from ...testing import make_name_without_spaces, TestCase
from ..testing import bind


def make_origin():
    # Create a new origin with BootSources and BootSource. The former refers
    # to the latter via the origin, hence why it must be bound.
    return bind(boot_sources.BootSources, boot_sources.BootSource)


class TestBootSource(TestCase):
    def test__string_representation_includes_url_keyring_info_only(self):
        source = boot_sources.BootSource(
            {
                "url": "http://images.maas.io/ephemeral-v3/daily/",
                "keyring_filename": (
                    "/usr/share/keyrings/ubuntu-cloudimage-keyring.gpg"
                ),
                "keyring_data": "",
            }
        )
        self.assertThat(
            repr(source),
            Equals(
                "<BootSource keyring_data=%(keyring_data)r "
                "keyring_filename=%(keyring_filename)r url=%(url)r>" % (source._data)
            ),
        )

    def test__read(self):
        source_id = random.randint(0, 100)
        url = "http://images.maas.io/ephemeral-v3/daily/"
        keyring_filename = "/usr/share/keyrings/ubuntu-cloudimage-keyring.gpg"

        BootSource = make_origin().BootSource
        BootSource._handler.read.return_value = {
            "id": source_id,
            "url": url,
            "keyring_filename": keyring_filename,
            "keyring_data": "",
        }

        source = BootSource.read(source_id)
        BootSource._handler.read.assert_called_once_with(id=source_id)
        self.assertThat(
            source,
            MatchesStructure.byEquality(
                id=source_id,
                url=url,
                keyring_filename=keyring_filename,
                keyring_data="",
            ),
        )

    def test__delete(self):
        source_id = random.randint(0, 100)

        BootSource = make_origin().BootSource
        source = BootSource(
            {
                "id": source_id,
                "url": "http://images.maas.io/ephemeral-v3/daily/",
                "keyring_filename": (
                    "/usr/share/keyrings/ubuntu-cloudimage-keyring.gpg"
                ),
                "keyring_data": "",
            }
        )

        source.delete()
        BootSource._handler.delete.assert_called_once_with(id=source_id)


class TestBootSources(TestCase):
    def test__read(self):
        BootSources = make_origin().BootSources
        BootSources._handler.read.return_value = [
            {"id": random.randint(0, 9)},
            {"id": random.randint(10, 19)},
        ]

        sources = BootSources.read()
        self.assertEquals(2, len(sources))

    def test__create_calls_create_with_keyring_filename(self):
        source_id = random.randint(0, 100)
        url = "http://images.maas.io/ephemeral-v3/daily/"
        keyring_filename = "/usr/share/keyrings/ubuntu-cloudimage-keyring.gpg"

        BootSources = make_origin().BootSources
        BootSources._handler.create.return_value = {
            "id": source_id,
            "url": url,
            "keyring_filename": keyring_filename,
            "keyring_data": "",
        }

        source = BootSources.create(url, keyring_filename=keyring_filename)
        BootSources._handler.create.assert_called_once_with(
            url=url, keyring_filename=keyring_filename, keyring_data=""
        )
        self.assertThat(
            source,
            MatchesStructure.byEquality(
                id=source_id,
                url=url,
                keyring_filename=keyring_filename,
                keyring_data="",
            ),
        )

    def test__create_calls_create_with_keyring_data(self):
        source_id = random.randint(0, 100)
        url = "http://images.maas.io/ephemeral-v3/daily/"
        keyring_data = make_name_without_spaces("data")

        BootSources = make_origin().BootSources
        BootSources._handler.create.return_value = {
            "id": source_id,
            "url": url,
            "keyring_filename": "",
            "keyring_data": keyring_data,
        }

        source = BootSources.create(url, keyring_data=keyring_data)
        BootSources._handler.create.assert_called_once_with(
            url=url, keyring_filename="", keyring_data=keyring_data
        )
        self.assertThat(
            source,
            MatchesStructure.byEquality(
                id=source_id, url=url, keyring_filename="", keyring_data=keyring_data
            ),
        )

    def test__create_calls_create_with_unsigned_url(self):
        source_id = random.randint(0, 100)
        url = "http://images.maas.io/ephemeral-v3/daily/streams/v1/index.json"

        BootSources = make_origin().BootSources
        BootSources._handler.create.return_value = {
            "id": source_id,
            "url": url,
            "keyring_filename": "",
            "keyring_data": "",
        }

        source = BootSources.create(url)
        BootSources._handler.create.assert_called_once_with(
            url=url, keyring_filename="", keyring_data=""
        )
        self.assertThat(
            source,
            MatchesStructure.byEquality(
                id=source_id, url=url, keyring_filename="", keyring_data=""
            ),
        )



"""Test for `maas.client.viscera.bcache_cache_sets`."""

import random

from testtools.matchers import Equals

from ..bcache_cache_sets import BcacheCacheSet, BcacheCacheSets
from ..bcaches import Bcache, Bcaches
from ..block_devices import BlockDevice, BlockDevices
from ..nodes import Node, Nodes
from ..partitions import Partition, Partitions

from ..testing import bind
from ...enum import BlockDeviceType, CacheMode
from ...testing import make_string_without_spaces, TestCase


def make_origin():
    """
    Create a new origin with required objects.
    """
    return bind(
        Bcaches,
        Bcache,
        BcacheCacheSets,
        BcacheCacheSet,
        BlockDevices,
        BlockDevice,
        Partitions,
        Partition,
        Nodes,
        Node,
    )


class TestBcaches(TestCase):
    def test__by_name(self):
        origin = make_origin()
        Bcaches, Bcache = origin.Bcaches, origin.Bcache
        system_id = make_string_without_spaces()
        bcache_names = [make_string_without_spaces() for _ in range(3)]
        bcaches_by_name = {
            name: Bcache(
                {
                    "system_id": system_id,
                    "id": random.randint(0, 100),
                    "name": name,
                    "backing_device": {
                        "id": random.randint(0, 100),
                        "type": BlockDeviceType.PHYSICAL.value,
                    },
                    "cache_set": {"id": random.randint(0, 100)},
                    "virtual_device": {
                        "id": random.randint(0, 100),
                        "type": BlockDeviceType.VIRTUAL.value,
                    },
                }
            )
            for name in bcache_names
        }
        bcaches = Bcaches([obj for _, obj in bcaches_by_name.items()])
        self.assertEqual(bcaches_by_name, bcaches.by_name)

    def test__get_by_name(self):
        origin = make_origin()
        Bcaches, Bcache = origin.Bcaches, origin.Bcache
        system_id = make_string_without_spaces()
        bcache_names = [make_string_without_spaces() for _ in range(3)]
        bcaches_by_name = {
            name: Bcache(
                {
                    "system_id": system_id,
                    "id": random.randint(0, 100),
                    "name": name,
                    "backing_device": {
                        "id": random.randint(0, 100),
                        "type": BlockDeviceType.PHYSICAL.value,
                    },
                    "cache_set": {"id": random.randint(0, 100)},
                    "virtual_device": {
                        "id": random.randint(0, 100),
                        "type": BlockDeviceType.VIRTUAL.value,
                    },
                }
            )
            for name in bcache_names
        }
        bcaches = Bcaches([obj for _, obj in bcaches_by_name.items()])
        name = bcache_names[0]
        self.assertEqual(bcaches_by_name[name], bcaches.get_by_name(name))

    def test__read_bad_node_type(self):
        Bcaches = make_origin().Bcaches
        error = self.assertRaises(TypeError, Bcaches.read, random.randint(0, 100))
        self.assertEquals("node must be a Node or str, not int", str(error))

    def test__read_with_system_id(self):
        Bcaches = make_origin().Bcaches
        system_id = make_string_without_spaces()
        bcaches = [
            {
                "system_id": system_id,
                "id": random.randint(0, 100),
                "name": make_string_without_spaces(),
                "backing_device": {
                    "id": random.randint(0, 100),
                    "type": BlockDeviceType.PHYSICAL.value,
                },
                "cache_set": {"id": random.randint(0, 100)},
                "virtual_device": {
                    "id": random.randint(0, 100),
                    "type": BlockDeviceType.VIRTUAL.value,
                },
            }
            for _ in range(3)
        ]
        Bcaches._handler.read.return_value = bcaches
        bcaches = Bcaches.read(system_id)
        self.assertThat(len(bcaches), Equals(3))
        Bcaches._handler.read.assert_called_once_with(system_id=system_id)

    def test__read_with_Node(self):
        origin = make_origin()
        Bcaches, Node = origin.Bcaches, origin.Node
        system_id = make_string_without_spaces()
        node = Node(system_id)
        bcaches = [
            {
                "system_id": system_id,
                "id": random.randint(0, 100),
                "name": make_string_without_spaces(),
                "backing_device": {
                    "id": random.randint(0, 100),
                    "type": BlockDeviceType.PHYSICAL.value,
                },
                "cache_set": {"id": random.randint(0, 100)},
                "virtual_device": {
                    "id": random.randint(0, 100),
                    "type": BlockDeviceType.VIRTUAL.value,
                },
            }
            for _ in range(3)
        ]
        Bcaches._handler.read.return_value = bcaches
        bcaches = Bcaches.read(node)
        self.assertThat(len(bcaches), Equals(3))
        Bcaches._handler.read.assert_called_once_with(system_id=system_id)

    def test__create_bad_node_type(self):
        origin = make_origin()
        Bcaches = origin.Bcaches
        error = self.assertRaises(
            TypeError,
            Bcaches.create,
            random.randint(0, 100),
            make_string_without_spaces(),
            origin.BlockDevice({}),
            random.randint(0, 100),
            CacheMode.WRITETHROUGH,
        )
        self.assertEquals("node must be a Node or str, not int", str(error))

    def test__create_bad_cache_device_type(self):
        Bcaches = make_origin().Bcaches
        error = self.assertRaises(
            TypeError,
            Bcaches.create,
            make_string_without_spaces(),
            make_string_without_spaces(),
            random.randint(0, 100),
            random.randint(0, 100),
            CacheMode.WRITETHROUGH,
        )
        self.assertEquals(
            "backing_device must be a BlockDevice or Partition, not int", str(error)
        )

    def test__create_bad_cache_set_type(self):
        origin = make_origin()
        Bcaches = origin.Bcaches
        error = self.assertRaises(
            TypeError,
            Bcaches.create,
            make_string_without_spaces(),
            make_string_without_spaces(),
            origin.BlockDevice({"id": random.randint(0, 100)}),
            make_string_without_spaces(),
            CacheMode.WRITETHROUGH,
        )
        self.assertEquals(
            "cache_set must be a BcacheCacheSet or int, not str", str(error)
        )

    def test__create_bad_cache_mode_type(self):
        origin = make_origin()
        Bcaches = origin.Bcaches
        error = self.assertRaises(
            TypeError,
            Bcaches.create,
            make_string_without_spaces(),
            make_string_without_spaces(),
            origin.BlockDevice({"id": random.randint(0, 100)}),
            random.randint(0, 100),
            "writethrough",
        )
        self.assertEquals("cache_mode must be a CacheMode, not str", str(error))

    def test__create_with_block_device(self):
        origin = make_origin()
        Bcaches = origin.Bcaches
        BlockDevice = origin.BlockDevice
        BcacheCacheSet = origin.BcacheCacheSet
        block_device = BlockDevice({"id": random.randint(0, 100)})
        cache_set = BcacheCacheSet({"id": random.randint(0, 100)})
        Bcaches._handler.create.return_value = {
            "id": random.randint(0, 100),
            "backing_device": block_device._data,
            "cache_set": cache_set._data,
        }
        name = make_string_without_spaces()
        system_id = make_string_without_spaces()
        Bcaches.create(system_id, name, block_device, cache_set, CacheMode.WRITEBACK)
        Bcaches._handler.create.assert_called_once_with(
            system_id=system_id,
            name=name,
            backing_device=block_device.id,
            cache_set=cache_set.id,
            cache_mode=CacheMode.WRITEBACK.value,
        )

    def test__create_with_partition(self):
        origin = make_origin()
        Bcaches = origin.Bcaches
        Partition = origin.Partition
        partition = Partition({"id": random.randint(0, 100)})
        cache_set_id = random.randint(0, 100)
        Bcaches._handler.create.return_value = {
            "id": random.randint(0, 100),
            "backing_device": partition._data,
            "cache_set": {"id": cache_set_id},
        }
        name = make_string_without_spaces()
        system_id = make_string_without_spaces()
        Bcaches.create(system_id, name, partition, cache_set_id, CacheMode.WRITEBACK)
        Bcaches._handler.create.assert_called_once_with(
            system_id=system_id,
            name=name,
            backing_partition=partition.id,
            cache_set=cache_set_id,
            cache_mode=CacheMode.WRITEBACK.value,
        )


class TestBcache(TestCase):
    def test__read_bad_node_type(self):
        Bcache = make_origin().Bcache
        error = self.assertRaises(
            TypeError, Bcache.read, random.randint(0, 100), random.randint(0, 100)
        )
        self.assertEquals("node must be a Node or str, not int", str(error))

    def test__read_with_system_id(self):
        Bcache = make_origin().Bcache
        system_id = make_string_without_spaces()
        bcache = {
            "system_id": system_id,
            "id": random.randint(0, 100),
            "backing_device": {
                "id": random.randint(0, 100),
                "type": BlockDeviceType.PHYSICAL.value,
            },
        }
        Bcache._handler.read.return_value = bcache
        Bcache.read(system_id, bcache["id"])
        Bcache._handler.read.assert_called_once_with(
            system_id=system_id, id=bcache["id"]
        )

    def test__read_with_Node(self):
        origin = make_origin()
        Bcache = origin.Bcache
        Node = origin.Node
        system_id = make_string_without_spaces()
        node = Node(system_id)
        bcache = {
            "system_id": system_id,
            "id": random.randint(0, 100),
            "backing_device": {
                "id": random.randint(0, 100),
                "type": BlockDeviceType.PHYSICAL.value,
            },
        }
        Bcache._handler.read.return_value = bcache
        Bcache.read(node, bcache["id"])
        Bcache._handler.read.assert_called_once_with(
            system_id=system_id, id=bcache["id"]
        )

    def test__delete(self):
        Bcache = make_origin().Bcache
        system_id = make_string_without_spaces()
        bcache = Bcache(
            {
                "system_id": system_id,
                "id": random.randint(0, 100),
                "backing_device": {
                    "id": random.randint(0, 100),
                    "type": BlockDeviceType.PHYSICAL.value,
                },
            }
        )
        Bcache._handler.delete.return_value = None
        bcache.delete()
        Bcache._handler.delete.assert_called_once_with(
            system_id=system_id, id=bcache.id
        )



"""Tests for `maas.client.flesh.devices`."""

from operator import itemgetter
import yaml

from .testing import TestCaseWithProfile
from .. import ArgumentParser, devices, tabular
from ...testing import make_name_without_spaces
from ...viscera.testing import bind
from ...viscera.devices import Device, Devices
from ...viscera.interfaces import Interface, Interfaces, InterfaceLink, InterfaceLinks
from ...viscera.users import User


def make_origin():
    """Make origin for devices."""
    return bind(
        Devices, Device, User, Interfaces, Interface, InterfaceLinks, InterfaceLink
    )


class TestDevices(TestCaseWithProfile):
    """Tests for `cmd_devices`."""

    def test_returns_table_with_devices(self):
        origin = make_origin()
        parser = ArgumentParser()
        devices_objs = [
            {
                "hostname": make_name_without_spaces(),
                "owner": make_name_without_spaces(),
                "interface_set": [
                    {"links": [{"ip_address": "192.168.122.1"}]},
                    {"links": [{"ip_address": "192.168.122.2"}]},
                    {"links": [{}]},
                ],
            },
            {
                "hostname": make_name_without_spaces(),
                "owner": make_name_without_spaces(),
                "interface_set": [
                    {"links": [{"ip_address": "192.168.122.10"}]},
                    {"links": [{"ip_address": "192.168.122.11"}]},
                    {"links": [{}]},
                ],
            },
        ]
        origin.Devices._handler.read.return_value = devices_objs
        cmd = devices.cmd_devices(parser)
        subparser = devices.cmd_devices.register(parser)
        options = subparser.parse_args([])
        output = yaml.safe_load(
            cmd.execute(origin, options, target=tabular.RenderTarget.yaml)
        )
        self.assertEquals(
            [
                {"name": "hostname", "title": "Hostname"},
                {"name": "owner", "title": "Owner"},
                {"name": "ip_addresses", "title": "IP addresses"},
            ],
            output["columns"],
        )
        devices_output = sorted(
            [
                {
                    "hostname": device["hostname"],
                    "owner": device["owner"] if device["owner"] else "(none)",
                    "ip_addresses": [
                        link["ip_address"]
                        for nic in device["interface_set"]
                        for link in nic["links"]
                        if link.get("ip_address")
                    ],
                }
                for device in devices_objs
            ],
            key=itemgetter("hostname"),
        )
        self.assertEquals(devices_output, output["data"])

    def test_calls_handler_with_hostnames(self):
        origin = make_origin()
        parser = ArgumentParser()
        origin.Devices._handler.read.return_value = []
        subparser = devices.cmd_devices.register(parser)
        cmd = devices.cmd_devices(parser)
        hostnames = [make_name_without_spaces() for _ in range(3)]
        options = subparser.parse_args(hostnames)
        cmd.execute(origin, options, target=tabular.RenderTarget.yaml)
        origin.Devices._handler.read.assert_called_once_with(hostname=hostnames)



"""Tests for `maas.client.viscera.tags`."""

from testtools.matchers import Equals, IsInstance, MatchesStructure

from .. import tags

from ..testing import bind
from ...testing import make_string_without_spaces, TestCase


def make_origin():
    """
    Create a new origin with Tag and Tags. The former
    refers to the latter via the origin, hence why it must be bound.
    """
    return bind(tags.Tags, tags.Tag)


class TestTags(TestCase):
    def test__tags_create(self):
        origin = make_origin()
        name = make_string_without_spaces()
        comment = make_string_without_spaces()
        origin.Tags._handler.create.return_value = {"name": name, "comment": comment}
        tag = origin.Tags.create(name=name, comment=comment)
        origin.Tags._handler.create.assert_called_once_with(name=name, comment=comment)
        self.assertThat(tag, IsInstance(origin.Tag))
        self.assertThat(tag, MatchesStructure.byEquality(name=name, comment=comment))

    def test__tags_create_without_comment(self):
        origin = make_origin()
        name = make_string_without_spaces()
        comment = ""
        origin.Tags._handler.create.return_value = {"name": name, "comment": comment}
        tag = origin.Tags.create(name=name)
        origin.Tags._handler.create.assert_called_once_with(name=name)
        self.assertThat(tag, IsInstance(origin.Tag))
        self.assertThat(tag, MatchesStructure.byEquality(name=name, comment=comment))

    def test__tags_read(self):
        """Tags.read() returns a list of tags."""
        Tags = make_origin().Tags
        tags = [
            {
                "name": make_string_without_spaces(),
                "comment": make_string_without_spaces(),
            }
            for _ in range(3)
        ]
        Tags._handler.read.return_value = tags
        tags = Tags.read()
        self.assertThat(len(tags), Equals(3))


class TestTag(TestCase):
    def test__tag_read(self):
        Tag = make_origin().Tag
        tag = {
            "name": make_string_without_spaces(),
            "comment": make_string_without_spaces(),
        }
        Tag._handler.read.return_value = tag
        self.assertThat(Tag.read(name=tag["name"]), Equals(Tag(tag)))
        Tag._handler.read.assert_called_once_with(name=tag["name"])

    def test__tag_delete(self):
        Tag = make_origin().Tag
        tag_name = make_string_without_spaces()
        tag = Tag({"name": tag_name, "comment": make_string_without_spaces()})
        tag.delete()
        Tag._handler.delete.assert_called_once_with(name=tag_name)



"""Commands for controllers."""

__all__ = ["register"]

import asyncio
from itertools import chain

from . import CommandError, OriginPagedTableCommand, tables
from ..enum import NodeType
from ..utils.maas_async import asynchronous


class cmd_controllers(OriginPagedTableCommand):
    """List controllers."""

    def __init__(self, parser):
        super(cmd_controllers, self).__init__(parser)
        parser.add_argument("hostname", nargs="*", help=("Hostname of the controller."))

    @asynchronous
    async def execute(self, origin, options, target):
        hostnames = None
        if options.hostname:
            hostnames = options.hostname

        controller_sets = await asyncio.gather(
            origin.RackControllers.read(hostnames=hostnames),
            origin.RegionControllers.read(hostnames=hostnames),
        )
        controllers = {
            controller.system_id: controller
            for controller in chain.from_iterable(controller_sets)
        }
        table = tables.ControllersTable()
        return table.render(target, controllers.values())


class cmd_controller(OriginPagedTableCommand):
    """Details of a controller."""

    def __init__(self, parser):
        super(cmd_controller, self).__init__(parser)
        parser.add_argument("hostname", nargs=1, help=("Hostname of the controller."))

    def execute(self, origin, options, target):
        nodes = origin.Nodes.read(hostnames=options.hostname)
        if len(nodes) == 0:
            raise CommandError("Unable to find controller %s." % options.hostname[0])
        node = nodes[0]
        if node.node_type == NodeType.RACK_CONTROLLER:
            table = tables.ControllerDetail()
            node = node.as_rack_controller()
        elif node.node_type == NodeType.REGION_CONTROLLER:
            table = tables.ControllerDetail()
            node = node.as_region_controller()
        elif node.node_type == NodeType.REGION_AND_RACK_CONTROLLER:
            table = tables.ControllerDetail()
            node = node.as_rack_controller()
        else:
            raise CommandError("Unable to find controller %s." % options.hostname[0])
        return table.render(target, node)


def register(parser):
    """Register commands with the given parser."""
    cmd_controllers.register(parser)
    cmd_controller.register(parser)



"""Objects for Bcaches."""

__all__ = ["Bcache", "Bcaches"]

from typing import Union

from . import ObjectField, ObjectFieldRelated, ObjectSet, ObjectType, to, check
from .nodes import Node
from .bcache_cache_sets import BcacheCacheSet
from .block_devices import BlockDevice
from .partitions import Partition
from .filesystem_groups import DeviceField, FilesystemGroup
from ..enum import CacheMode


class BcacheType(ObjectType):
    """Metaclass for `Bcache`."""

    async def read(cls, node, id):
        """Get `Bcache` by `id`."""
        if isinstance(node, str):
            system_id = node
        elif isinstance(node, Node):
            system_id = node.system_id
        else:
            raise TypeError("node must be a Node or str, not %s" % type(node).__name__)
        return cls(await cls._handler.read(system_id=system_id, id=id))


class Bcache(FilesystemGroup, metaclass=BcacheType):
    """A Bcache on a machine."""

    cache_mode = ObjectField.Checked("cache_mode", to(CacheMode), check(CacheMode))
    uuid = ObjectField.Checked("uuid", check(str), check(str))

    backing_device = DeviceField("backing_device")
    cache_set = ObjectFieldRelated("cache_set", "BcacheCacheSet", reverse=None)
    virtual_device = ObjectFieldRelated(
        "virtual_device", "BlockDevice", reverse=None, readonly=True
    )

    def __repr__(self):
        return super(Bcache, self).__repr__(
            fields={"name", "cache_mode", "size", "backing_device"}
        )

    async def delete(self):
        """Delete this Bcache."""
        await self._handler.delete(system_id=self.node.system_id, id=self.id)


class BcachesType(ObjectType):
    """Metaclass for `Bcaches`."""

    async def read(cls, node):
        """Get list of `Bcache`'s for `node`."""
        if isinstance(node, str):
            system_id = node
        elif isinstance(node, Node):
            system_id = node.system_id
        else:
            raise TypeError("node must be a Node or str, not %s" % type(node).__name__)
        data = await cls._handler.read(system_id=system_id)
        return cls(
            cls._object(item, local_data={"node_system_id": system_id}) for item in data
        )

    async def create(
        cls,
        node: Union[Node, str],
        name: str,
        backing_device: Union[BlockDevice, Partition],
        cache_set: Union[BcacheCacheSet, int],
        cache_mode: CacheMode,
        *,
        uuid: str = None
    ):
        """
        Create a Bcache on a Node.

        :param node: Node to create the interface on.
        :type node: `Node` or `str`
        :param name: Name of the Bcache.
        :type name: `str`
        :param backing_device: Either a block device or partition to create
            the Bcache from.
        :type backing_device: `BlockDevice` or `Partition`
        :param cache_set: Bcache cache set to use in front of backing device.
        :type cache_set: `BcacheCacheSet` or `int`
        :param cache_mode: Caching mode to use for this device.
        :type cache_mode: `CacheMode`
        :type backing_device: `BlockDevice` or `Partition`
        :param uuid: The UUID for the Bcache (optional).
        :type uuid: `str`
        """
        params = {"name": name}
        if isinstance(node, str):
            params["system_id"] = node
        elif isinstance(node, Node):
            params["system_id"] = node.system_id
        else:
            raise TypeError(
                "node must be a Node or str, not %s" % (type(node).__name__)
            )

        if isinstance(backing_device, BlockDevice):
            params["backing_device"] = backing_device.id
        elif isinstance(backing_device, Partition):
            params["backing_partition"] = backing_device.id
        else:
            raise TypeError(
                "backing_device must be a BlockDevice or Partition, "
                "not %s" % type(backing_device).__name__
            )

        if isinstance(cache_set, BcacheCacheSet):
            params["cache_set"] = cache_set.id
        elif isinstance(cache_set, int):
            params["cache_set"] = cache_set
        else:
            raise TypeError(
                "cache_set must be a BcacheCacheSet or int, "
                "not %s" % type(cache_set).__name__
            )

        if isinstance(cache_mode, CacheMode):
            params["cache_mode"] = cache_mode.value
        else:
            raise TypeError(
                "cache_mode must be a CacheMode, " "not %s" % type(cache_mode).__name__
            )

        if uuid is not None:
            params["uuid"] = uuid
        return cls._object(await cls._handler.create(**params))


class Bcaches(ObjectSet, metaclass=BcachesType):
    """The set of Bcaches on a machine."""

    @property
    def by_name(self):
        """Return mapping of name to `Bcache`."""
        return {bcache.name: bcache for bcache in self}

    def get_by_name(self, name):
        """Return a `Bcache` by its name."""
        return self.by_name[name]



# Copyright 2016-2017 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for `maas.client.utils.auth`."""

import sys
from unittest.mock import ANY, sentinel

from .. import auth
from ...testing import TestCase
from ..testing import make_Credentials


class TestAuth(TestCase):
    def test_try_getpass(self):
        getpass = self.patch(auth, "getpass")
        getpass.return_value = sentinel.credentials
        self.assertIs(sentinel.credentials, auth.try_getpass(sentinel.prompt))
        getpass.assert_called_once_with(sentinel.prompt)

    def test_try_getpass_eof(self):
        getpass = self.patch(auth, "getpass")
        getpass.side_effect = EOFError
        self.assertIsNone(auth.try_getpass(sentinel.prompt))
        getpass.assert_called_once_with(sentinel.prompt)

    def test_obtain_credentials_from_stdin(self):
        # When "-" is passed to obtain_credentials, it reads credentials from
        # stdin, trims whitespace, and converts it into a 3-tuple of creds.
        credentials = make_Credentials()
        stdin = self.patch(sys, "stdin")
        stdin.readline.return_value = str(credentials) + "\n"
        self.assertEqual(credentials, auth.obtain_credentials("-"))
        stdin.readline.assert_called_once_with()

    def test_obtain_credentials_via_getpass(self):
        # When None is passed to obtain_credentials, it attempts to obtain
        # credentials via getpass, then converts it into a 3-tuple of creds.
        credentials = make_Credentials()
        getpass = self.patch(auth, "getpass")
        getpass.return_value = str(credentials)
        self.assertEqual(credentials, auth.obtain_credentials(None))
        getpass.assert_called_once_with(ANY)

    def test_obtain_credentials_empty(self):
        # If the entered credentials are empty or only whitespace,
        # obtain_credentials returns None.
        getpass = self.patch(auth, "getpass")
        getpass.return_value = None
        self.assertEqual(None, auth.obtain_credentials(None))
        getpass.assert_called_once_with(ANY)



"""Test for `maas.client.viscera.boot_resources`."""

import hashlib
from http import HTTPStatus
import io
import random
from unittest.mock import call, MagicMock, sentinel

import aiohttp
from aiohttp.test_utils import make_mocked_coro
from testtools.matchers import Equals, MatchesDict, MatchesStructure

from .. import boot_resources
from ...testing import (
    AsyncCallableMock,
    AsyncContextMock,
    make_name_without_spaces,
    make_string,
    pick_bool,
    TestCase,
)
from ..testing import bind


def make_origin():
    # Create a new origin with BootResources and BootResource. The former
    # refers to the latter via the origin, hence why it must be bound.
    return bind(boot_resources.BootResources, boot_resources.BootResource)


class TestBootResource(TestCase):
    def test__string_representation_includes_type_name_architecture(self):
        source = boot_resources.BootResource(
            {
                "id": random.randint(0, 100),
                "type": "Synced",
                "name": "ubuntu/xenial",
                "architecture": "amd64/ga-16.04",
                "subarches": "generic,ga-16.04",
            }
        )
        self.assertThat(
            repr(source),
            Equals(
                "<BootResource architecture=%(architecture)r name=%(name)r "
                "type=%(type)r>" % (source._data)
            ),
        )

    def test__read(self):
        resource_id = random.randint(0, 100)
        rtype = random.choice(["Synced", "Uploaded", "Generated"])
        name = "%s/%s" % (
            make_name_without_spaces("os"),
            make_name_without_spaces("release"),
        )
        architecture = "%s/%s" % (
            make_name_without_spaces("arch"),
            make_name_without_spaces("subarch"),
        )
        subarches = ",".join(make_name_without_spaces("subarch") for _ in range(3))
        sets = {}
        for _ in range(3):
            version = make_name_without_spaces("version")
            files = {}
            for _ in range(3):
                filename = make_name_without_spaces("filename")
                files[filename] = {
                    "filename": filename,
                    "filetype": make_name_without_spaces("filetype"),
                    "size": random.randint(1000, 10000),
                    "sha256": make_name_without_spaces("sha256"),
                    "complete": pick_bool(),
                }
            sets[version] = {
                "version": version,
                "size": random.randint(1000, 10000),
                "label": make_name_without_spaces("label"),
                "complete": pick_bool(),
                "files": files,
            }

        BootResource = make_origin().BootResource
        BootResource._handler.read.return_value = {
            "id": resource_id,
            "type": rtype,
            "name": name,
            "architecture": architecture,
            "subarches": subarches,
            "sets": sets,
        }

        resource = BootResource.read(resource_id)
        BootResource._handler.read.assert_called_once_with(id=resource_id)
        self.assertThat(
            resource,
            MatchesStructure(
                id=Equals(resource_id),
                type=Equals(rtype),
                name=Equals(name),
                architecture=Equals(architecture),
                subarches=Equals(subarches),
                sets=MatchesDict(
                    {
                        version: MatchesStructure(
                            version=Equals(version),
                            size=Equals(rset["size"]),
                            label=Equals(rset["label"]),
                            complete=Equals(rset["complete"]),
                            files=MatchesDict(
                                {
                                    filename: MatchesStructure(
                                        filename=Equals(filename),
                                        filetype=Equals(rfile["filetype"]),
                                        size=Equals(rfile["size"]),
                                        sha256=Equals(rfile["sha256"]),
                                        complete=Equals(rfile["complete"]),
                                    )
                                    for filename, rfile in rset["files"].items()
                                }
                            ),
                        )
                        for version, rset in sets.items()
                    }
                ),
            ),
        )

    def test__delete(self):
        resource_id = random.randint(0, 100)

        BootResource = make_origin().BootResource
        resource = BootResource(
            {
                "id": resource_id,
                "type": "Synced",
                "name": "ubuntu/xenial",
                "architecture": "amd64/ga-16.04",
                "subarches": "generic,ga-16.04",
            }
        )

        resource.delete()
        BootResource._handler.delete.assert_called_once_with(id=resource_id)


class TestBootResources(TestCase):
    def test__read(self):
        BootResources = make_origin().BootResources
        BootResources._handler.read.return_value = [
            {"id": random.randint(0, 9)},
            {"id": random.randint(10, 19)},
        ]

        resources = BootResources.read()
        self.assertEquals(2, len(resources))

    def test__start_import(self):
        BootResources = make_origin().BootResources
        import_action = getattr(BootResources._handler, "import")
        import_action.return_value = sentinel.result

        self.assertEquals(sentinel.result, BootResources.start_import())
        import_action.assert_called_once_with()

    def test__stop_import(self):
        BootResources = make_origin().BootResources
        BootResources._handler.stop_import.return_value = sentinel.result

        self.assertEquals(sentinel.result, BootResources.stop_import())
        BootResources._handler.stop_import.assert_called_once_with()

    def test__create_raises_ValueError_when_name_missing_slash(self):
        BootResources = make_origin().BootResources

        buf = io.BytesIO(b"")
        error = self.assertRaises(ValueError, BootResources.create, "", "", buf)
        self.assertEquals("name must be in format os/release; missing '/'", str(error))

    def test__create_raises_ValueError_when_architecture_missing_slash(self):
        BootResources = make_origin().BootResources

        buf = io.BytesIO(b"")
        error = self.assertRaises(
            ValueError, BootResources.create, "os/release", "", buf
        )
        self.assertEquals(
            "architecture must be in format arch/subarch; missing '/'", str(error)
        )

    def test__create_raises_ValueError_when_content_cannot_be_read(self):
        BootResources = make_origin().BootResources

        buf = io.BytesIO(b"")
        self.patch(buf, "readable").return_value = False
        error = self.assertRaises(
            ValueError, BootResources.create, "os/release", "arch/subarch", buf
        )
        self.assertEquals("content must be readable", str(error))

    def test__create_raises_ValueError_when_content_cannot_seek(self):
        BootResources = make_origin().BootResources

        buf = io.BytesIO(b"")
        self.patch(buf, "seekable").return_value = False
        error = self.assertRaises(
            ValueError, BootResources.create, "os/release", "arch/subarch", buf
        )
        self.assertEquals("content must be seekable", str(error))

    def test__create_raises_ValueError_when_chunk_size_is_zero(self):
        BootResources = make_origin().BootResources

        buf = io.BytesIO(b"")
        error = self.assertRaises(
            ValueError,
            BootResources.create,
            "os/release",
            "arch/subarch",
            buf,
            chunk_size=0,
        )
        self.assertEquals("chunk_size must be greater than 0, not 0", str(error))

    def test__create_raises_ValueError_when_chunk_size_is_less_than_zero(self):
        BootResources = make_origin().BootResources

        buf = io.BytesIO(b"")
        error = self.assertRaises(
            ValueError,
            BootResources.create,
            "os/release",
            "arch/subarch",
            buf,
            chunk_size=-1,
        )
        self.assertEquals("chunk_size must be greater than 0, not -1", str(error))

    def test__create_calls_create_on_handler_does_nothing_if_complete(self):
        resource_id = random.randint(0, 100)
        name = "%s/%s" % (
            make_name_without_spaces("os"),
            make_name_without_spaces("release"),
        )
        architecture = "%s/%s" % (
            make_name_without_spaces("arch"),
            make_name_without_spaces("subarch"),
        )
        title = make_name_without_spaces("title")
        filetype = random.choice(
            [
                boot_resources.BootResourceFileType.TGZ,
                boot_resources.BootResourceFileType.DDTGZ,
            ]
        )

        data = make_string().encode("ascii")
        sha256 = hashlib.sha256()
        sha256.update(data)
        sha256 = sha256.hexdigest()
        size = len(data)
        buf = io.BytesIO(data)

        BootResources = make_origin().BootResources
        BootResources._handler.create.return_value = {
            "id": resource_id,
            "type": "Uploaded",
            "name": name,
            "architecture": architecture,
            "sets": {
                "20161026": {
                    "version": "20161026",
                    "size": size,
                    "label": "uploaded",
                    "complete": True,
                    "files": {
                        "root-tgz": {
                            "filename": "root-tgz",
                            "filetype": "root-dd",
                            "size": size,
                            "sha256": sha256,
                            "complete": True,
                        }
                    },
                }
            },
        }
        mock_upload_chunks = self.patch(BootResources, "_upload_chunks")

        resource = BootResources.create(
            name, architecture, buf, title=title, filetype=filetype
        )
        self.assertThat(
            resource,
            MatchesStructure.byEquality(
                id=resource_id, type="Uploaded", name=name, architecture=architecture
            ),
        )
        self.assertFalse(mock_upload_chunks.called)

    def test__create_uploads_in_chunks_and_reloads_resource(self):
        resource_id = random.randint(0, 100)
        name = "%s/%s" % (
            make_name_without_spaces("os"),
            make_name_without_spaces("release"),
        )
        architecture = "%s/%s" % (
            make_name_without_spaces("arch"),
            make_name_without_spaces("subarch"),
        )
        title = make_name_without_spaces("title")
        filetype = random.choice(
            [
                boot_resources.BootResourceFileType.TGZ,
                boot_resources.BootResourceFileType.DDTGZ,
            ]
        )
        upload_uri = "/MAAS/api/2.0/boot-resources/%d/upload/1" % resource_id

        # Make chunks and upload in pieces of 4, where the last piece is
        # smaller than chunk size.
        chunk_size = random.randint(10, 100)
        data = make_string(int(chunk_size * 3.25)).encode("ascii")
        sha256 = hashlib.sha256()
        sha256.update(data)
        sha256 = sha256.hexdigest()
        size = len(data)
        buf = io.BytesIO(data)

        # Mock the handler calls. BootResource.read will be called after the
        # upload is complete to get the updated object.
        origin = make_origin()
        BootResources = origin.BootResources
        BootResource = origin.BootResource
        BootResources._handler.uri = (
            "http://localhost:5240/MAAS/api/2.0/boot-resources/"
        )
        BootResources._handler.create.return_value = {
            "id": resource_id,
            "type": "Uploaded",
            "name": name,
            "architecture": architecture,
            "sets": {
                "20161026": {
                    "version": "20161026",
                    "size": size,
                    "label": "uploaded",
                    "complete": False,
                    "files": {
                        "root-tgz": {
                            "filename": "root-tgz",
                            "filetype": "root-dd",
                            "size": size,
                            "sha256": sha256,
                            "complete": False,
                            "upload_uri": upload_uri,
                        }
                    },
                }
            },
        }
        BootResource._handler.read.return_value = {
            "id": resource_id,
            "type": "Uploaded",
            "name": name,
            "architecture": architecture,
            "sets": {
                "20161026": {
                    "version": "20161026",
                    "size": size,
                    "label": "uploaded",
                    "complete": True,
                    "files": {
                        "root-tgz": {
                            "filename": "root-tgz",
                            "filetype": "root-dd",
                            "size": size,
                            "sha256": sha256,
                            "complete": True,
                        }
                    },
                }
            },
        }

        # Mock signing. Test checks that its actually called.
        mock_sign = self.patch(boot_resources.utils, "sign")

        # Mock ClientSession.put as the create does PUT directly to the API.
        response = AsyncContextMock(spec=aiohttp.ClientResponse)
        response.status = HTTPStatus.OK.value

        put = AsyncCallableMock(return_value=response)
        self.patch(boot_resources.aiohttp.ClientSession, "put", put)

        # Progress handler called on each chunk.
        progress_handler = MagicMock()

        # Create and upload the resource.
        resource = BootResources.create(
            name,
            architecture,
            buf,
            title=title,
            filetype=filetype,
            chunk_size=chunk_size,
            progress_callback=progress_handler,
        )

        # Check that returned resource is correct and updated.
        self.assertThat(
            resource,
            MatchesStructure.byEquality(
                id=resource_id, type="Uploaded", name=name, architecture=architecture
            ),
        )
        self.assertTrue(resource.sets["20161026"].complete)

        # Check that the request was signed.
        self.assertTrue(mock_sign.called)

        # Check that the PUT was called for each chunk.
        calls = [
            call(
                "http://localhost:5240/MAAS/api/2.0/"
                "boot-resources/%d/upload/1" % resource_id,
                data=data[0 + i : chunk_size + i],
                headers={
                    "Content-Type": "application/octet-stream",
                    "Content-Length": str(len(data[0 + i : chunk_size + i])),
                },
            )
            for i in range(0, len(data), chunk_size)
        ]
        self.assertEquals(calls, put.call_args_list)

        # Check that progress handler was called on each chunk.
        calls = [
            call(len(data[: chunk_size + i]) / len(data))
            for i in range(0, len(data), chunk_size)
        ]
        self.assertEquals(calls, progress_handler.call_args_list)

    def test__create_raises_CallError_on_chunk_upload_failure(self):
        resource_id = random.randint(0, 100)
        name = "%s/%s" % (
            make_name_without_spaces("os"),
            make_name_without_spaces("release"),
        )
        architecture = "%s/%s" % (
            make_name_without_spaces("arch"),
            make_name_without_spaces("subarch"),
        )
        title = make_name_without_spaces("title")
        filetype = random.choice(
            [
                boot_resources.BootResourceFileType.TGZ,
                boot_resources.BootResourceFileType.DDTGZ,
            ]
        )
        upload_uri = "/MAAS/api/2.0/boot-resources/%d/upload/1" % resource_id

        # Make chunks and upload in pieces of 4, where the last piece is
        # smaller than chunk size.
        chunk_size = random.randint(10, 100)
        data = make_string(int(chunk_size * 3.25)).encode("ascii")
        sha256 = hashlib.sha256()
        sha256.update(data)
        sha256 = sha256.hexdigest()
        size = len(data)
        buf = io.BytesIO(data)

        # Mock the handler calls. BootResource.read will be called after the
        # upload is complete to get the updated object.
        origin = make_origin()
        BootResources = origin.BootResources
        BootResources._handler.uri = (
            "http://localhost:5240/MAAS/api/2.0/boot-resources/"
        )
        BootResources._handler.path = "/MAAS/api/2.0/boot-resources/"
        BootResources._handler.create.return_value = {
            "id": resource_id,
            "type": "Uploaded",
            "name": name,
            "architecture": architecture,
            "sets": {
                "20161026": {
                    "version": "20161026",
                    "size": size,
                    "label": "uploaded",
                    "complete": False,
                    "files": {
                        "root-tgz": {
                            "filename": "root-tgz",
                            "filetype": "root-dd",
                            "size": size,
                            "sha256": sha256,
                            "complete": False,
                            "upload_uri": upload_uri,
                        }
                    },
                }
            },
        }

        # Mock signing. Test checks that its actually called.
        mock_sign = self.patch(boot_resources.utils, "sign")

        # Mock ClientSession.put as the create does PUT directly to the API.
        response = AsyncContextMock(spec=aiohttp.ClientResponse)
        response.status = HTTPStatus.INTERNAL_SERVER_ERROR.value
        response.read = make_mocked_coro(b"Error")

        put = AsyncCallableMock(return_value=response)
        self.patch(boot_resources.aiohttp.ClientSession, "put", put)

        self.assertRaises(
            boot_resources.CallError,
            BootResources.create,
            name,
            architecture,
            buf,
            title=title,
            filetype=filetype,
            chunk_size=chunk_size,
        )

        # Check that the request was signed.
        self.assertTrue(mock_sign.called)



"""Tests for `maas.client.viscera`."""

from random import randrange, randint
from unittest.mock import call, Mock, sentinel

from testtools.matchers import (
    Contains,
    ContainsAll,
    Equals,
    HasLength,
    Is,
    IsInstance,
    MatchesStructure,
    Not,
)

from .. import (
    check,
    dir_class,
    dir_instance,
    Disabled,
    Object,
    ObjectBasics,
    ObjectField,
    ObjectFieldRelated,
    ObjectFieldRelatedSet,
    ObjectMethod,
    ObjectSet,
    ObjectType,
    OriginBase,
)
from ... import bones, viscera
from ...errors import ObjectNotLoaded
from ...testing import AsyncCallableMock, make_name, make_name_without_spaces, TestCase
from ...utils.tests.test_profiles import make_profile


class TestDirClass(TestCase):
    """Tests for `dir_class`."""

    def test__includes_ObjectMethod_descriptors_with_class_methods(self):
        class Example:
            attribute = ObjectMethod(sentinel.function)

        self.assertThat(list(dir_class(Example)), Contains("attribute"))

    def test__excludes_ObjectMethod_descriptors_without_class_methods(self):
        class Example:
            attribute = ObjectMethod()

        self.assertThat(list(dir_class(Example)), Not(Contains("attribute")))

    def test__excludes_Disabled_class_descriptors(self):
        class Example:
            attribute = Disabled("foobar")

        self.assertThat(list(dir_class(Example)), Not(Contains("attribute")))

    def test__includes_other_class_attributes(self):
        class Example:
            alice = "is the first"
            bob = lambda self: "just bob"
            carol = classmethod(lambda cls: "carol")
            dave = property(lambda self: "dave or david")
            erin = staticmethod(lambda: "or eve?")

        self.assertThat(
            list(dir_class(Example)),
            ContainsAll(["alice", "bob", "carol", "dave", "erin"]),
        )

    def test__excludes_mro_metaclass_method(self):
        class Example:
            """Example class."""

        self.assertThat(list(dir_class(Example)), Not(Contains("mro")))

    def test__excludes_Disabled_metaclass_descriptors(self):
        class ExampleType(type):
            attribute = Disabled("foobar")

        class Example(metaclass=ExampleType):
            """Example class with metaclass."""

        self.assertThat(list(dir_class(Example)), Not(Contains("attribute")))

    def test__includes_superclass_attributes(self):
        class ExampleBase:
            alice = "is the first"
            bob = lambda self: "just bob"

        class Example(ExampleBase):
            carol = classmethod(lambda cls: "carol")
            dave = property(lambda self: "dave or david")
            erin = staticmethod(lambda: "or eve?")

        self.assertThat(
            list(dir_class(Example)),
            ContainsAll(["alice", "bob", "carol", "dave", "erin"]),
        )


class TestDirInstance(TestCase):
    """Tests for `dir_instance`."""

    def test__includes_ObjectMethod_descriptors_with_instance_methods(self):
        class Example:
            attribute = ObjectMethod()
            attribute.instancemethod(sentinel.function)

        example = Example()

        self.assertThat(list(dir_instance(example)), Contains("attribute"))

    def test__excludes_ObjectMethod_descriptors_without_instance_methods(self):
        class Example:
            attribute = ObjectMethod()
            attribute.classmethod(sentinel.function)

        example = Example()

        self.assertThat(list(dir_instance(example)), Not(Contains("attribute")))

    def test__excludes_Disabled_class_descriptors(self):
        class Example:
            attribute = Disabled("foobar")

        example = Example()

        self.assertThat(list(dir_instance(example)), Not(Contains("attribute")))

    def test__excludes_class_methods(self):
        class Example:
            carol = classmethod(lambda cls: "carol")

        example = Example()

        self.assertThat(list(dir_instance(example)), Not(Contains("carol")))

    def test__excludes_static_methods(self):
        class Example:
            steve = staticmethod(lambda: "or eve?")

        example = Example()

        self.assertThat(list(dir_instance(example)), Not(Contains("steve")))

    def test__includes_other_class_attributes(self):
        class Example:
            alice = "is the first"
            bob = lambda self: "just bob"
            dave = property(lambda self: "or david")

        example = Example()

        self.assertThat(
            list(dir_instance(example)), ContainsAll(["alice", "bob", "dave"])
        )

    def test__excludes_instance_attributes(self):
        # In a bit of a departure, dir_instance(foo) will NOT return instance
        # attributes of foo. This is because object attributes in viscera
        # should be defined using descriptors (which are class attributes).

        class Example:
            """Example class."""

        example = Example()
        example.alice = 123

        self.assertThat(list(dir_instance(example)), Not(Contains(["alice"])))


class TestObjectType(TestCase):
    """Tests for `ObjectType`."""

    def test__classes_always_have_slots_defined(self):
        class WithoutSlots(metaclass=ObjectType):
            """A class WITHOUT __slots__ defined explicitly."""

        self.assertThat(WithoutSlots.__slots__, Equals(()))
        self.assertRaises(AttributeError, getattr, WithoutSlots(), "__dict__")

        class WithSlots(metaclass=ObjectType):
            """A class WITH __slots__ defined explicitly."""

            __slots__ = "a", "b"

        self.assertThat(WithSlots.__slots__, Equals(("a", "b")))
        self.assertRaises(AttributeError, getattr, WithSlots(), "__dict__")

    def test__uses_dir_class(self):
        class Dummy(metaclass=ObjectType):
            """Does nothing; just a stand-in."""

        dir_class = self.patch(viscera, "dir_class")
        dir_class.return_value = iter([sentinel.name])

        self.assertThat(dir(Dummy), Equals([sentinel.name]))


class TestObjectBasics(TestCase):
    """Tests for `ObjectBasics`."""

    def test__defines_slots(self):
        self.assertThat(ObjectBasics.__slots__, Equals(()))

    def test__uses_dir_instance(self):
        dir_instance = self.patch(viscera, "dir_instance")
        dir_instance.return_value = iter([sentinel.name])

        self.assertThat(dir(ObjectBasics()), Equals([sentinel.name]))

    def test__stringification_returns_qualified_class_name(self):
        self.assertThat(str(ObjectBasics()), Equals(ObjectBasics.__qualname__))


class TestObject(TestCase):
    """Tests for `Object`."""

    def test__defines_slots(self):
        self.assertThat(
            Object.__slots__,
            Equals(("__weakref__", "_data", "_orig_data", "_changed_data", "_loaded")),
        )

    def test__inherits_ObjectBasics(self):
        self.assertThat(Object.__mro__, Contains(ObjectBasics))

    def test__init_sets__data_and_loaded(self):
        data = {"alice": make_name_without_spaces("alice")}
        self.assertThat(Object(data)._data, Equals(data))
        self.assertTrue(Object(data)._loaded)

    def test__init_insists_on_mapping_when_no_pk(self):
        error = self.assertRaises(TypeError, Object, ["some", "items"])
        self.assertThat(str(error), Equals("data must be a mapping, not list"))

    def test__init_insists_on_complete_data(self):
        data = {"alice": make_name_without_spaces("alice"), "__incomplete__": True}
        error = self.assertRaises(ValueError, Object, data)
        self.assertThat(
            str(error),
            Equals("data cannot be incomplete without any primary keys defined"),
        )

    def test__init_takes_pk_when_defined(self):
        object_type = type("PKObject", (Object,), {"pk": ObjectField("pk_d", pk=True)})
        object_pk = randint(0, 20)
        object_a = object_type(object_pk)
        self.assertThat(object_a._data, Equals({"pk_d": object_pk}))
        self.assertThat(object_a.pk, Equals(object_pk))
        self.assertFalse(object_a._loaded)

    def test__init_takes_pk_in_mapping_when_defined(self):
        object_type = type("PKObject", (Object,), {"pk": ObjectField("pk_d", pk=True)})
        object_pk = randint(0, 20)
        object_a = object_type({"pk_d": object_pk, "__incomplete__": True})
        self.assertThat(object_a._data, Equals({"pk_d": object_pk}))
        self.assertThat(object_a.pk, Equals(object_pk))
        self.assertFalse(object_a._loaded)

    def test__init_takes_alt_pk_in_mapping_when_defined(self):
        object_type = type(
            "PKObject",
            (Object,),
            {
                "pk": ObjectField("pk_d", pk=True),
                "alt_pk": ObjectField("alt_pk_d", alt_pk=True),
            },
        )
        object_alt_pk = randint(0, 20)
        object_a = object_type({"alt_pk_d": object_alt_pk, "__incomplete__": True})
        self.assertThat(object_a._data, Equals({"alt_pk_d": object_alt_pk}))
        self.assertThat(object_a.alt_pk, Equals(object_alt_pk))
        self.assertFalse(object_a._loaded)

    def test__init_validates_pk_when_defined(self):
        object_type = type(
            "PKObject",
            (Object,),
            {"pk": ObjectField.Checked("pk_d", check(int), pk=True)},
        )
        error = self.assertRaises(TypeError, object_type, "not int")
        self.assertThat(str(error), Equals("'not int' is not of type %r" % int))

    def test__init_validates_pk_in_mapping_when_defined(self):
        object_type = type(
            "PKObject",
            (Object,),
            {"pk": ObjectField.Checked("pk_d", check(int), pk=True)},
        )
        error = self.assertRaises(
            TypeError, object_type, {"pk_d": "not int", "__incomplete__": True}
        )
        self.assertThat(str(error), Equals("'not int' is not of type %r" % int))

    def test__init_validates_alt_pk_in_mapping_when_defined(self):
        object_type = type(
            "PKObject",
            (Object,),
            {
                "pk": ObjectField.Checked("pk_d", check(int), pk=True),
                "alt_pk": ObjectField.Checked("alt_pk_d", check(int), alt_pk=True),
            },
        )
        error = self.assertRaises(
            TypeError, object_type, {"alt_pk_d": "not int", "__incomplete__": True}
        )
        self.assertThat(str(error), Equals("'not int' is not of type %r" % int))

    def test__init_doesnt_allow_multiple_pk_True(self):
        object_type = type(
            "PKObject",
            (Object,),
            {
                "pk1": ObjectField.Checked("pk_1", check(int), pk=True),
                "pk2": ObjectField.Checked("pk_2", check(int), pk=True),
            },
        )
        error = self.assertRaises(AttributeError, object_type, [0, 1])
        self.assertThat(
            str(error),
            Equals("more than one field is marked as unique " "primary key: pk1, pk2"),
        )

    def test__init_allows_mapping_when_multiple_pks(self):
        object_type = type(
            "PKObject",
            (Object,),
            {
                "pk1": ObjectField.Checked("pk_1", check(int), pk=0),
                "pk2": ObjectField.Checked("pk_2", check(int), pk=1),
            },
        )
        object_pk_1 = randint(0, 20)
        object_pk_2 = randint(0, 20)
        object_a = object_type(
            {"pk_1": object_pk_1, "pk_2": object_pk_2, "__incomplete__": True}
        )
        self.assertThat(
            object_a._data, Equals({"pk_1": object_pk_1, "pk_2": object_pk_2})
        )
        self.assertThat(object_a.pk1, Equals(object_pk_1))
        self.assertThat(object_a.pk2, Equals(object_pk_2))
        self.assertFalse(object_a._loaded)

    def test__init_allows_mapping_when_multiple_alt_pks(self):
        object_type = type(
            "PKObject",
            (Object,),
            {
                "pk1": ObjectField.Checked("pk_1", check(int), pk=0),
                "pk2": ObjectField.Checked("pk_2", check(int), pk=1),
                "alt_pk1": ObjectField.Checked("alt_pk_1", check(int), alt_pk=0),
                "alt_pk2": ObjectField.Checked("alt_pk_2", check(int), alt_pk=1),
            },
        )
        object_alt_pk_1 = randint(0, 20)
        object_alt_pk_2 = randint(0, 20)
        object_a = object_type(
            {
                "alt_pk_1": object_alt_pk_1,
                "alt_pk_2": object_alt_pk_2,
                "__incomplete__": True,
            }
        )
        self.assertThat(
            object_a._data,
            Equals({"alt_pk_1": object_alt_pk_1, "alt_pk_2": object_alt_pk_2}),
        )
        self.assertThat(object_a.alt_pk1, Equals(object_alt_pk_1))
        self.assertThat(object_a.alt_pk2, Equals(object_alt_pk_2))
        self.assertFalse(object_a._loaded)

    def test__init_allows_sequence_when_multiple_pks(self):
        object_type = type(
            "PKObject",
            (Object,),
            {
                "pk1": ObjectField.Checked("pk_1", check(int), pk=0),
                "pk2": ObjectField.Checked("pk_2", check(int), pk=1),
            },
        )
        object_pk_1 = randint(0, 20)
        object_pk_2 = randint(0, 20)
        object_a = object_type([object_pk_1, object_pk_2])
        self.assertThat(
            object_a._data, Equals({"pk_1": object_pk_1, "pk_2": object_pk_2})
        )
        self.assertThat(object_a.pk1, Equals(object_pk_1))
        self.assertThat(object_a.pk2, Equals(object_pk_2))
        self.assertFalse(object_a._loaded)

    def test__init_requires_mapping_or_sequence_when_multiple_pks(self):
        object_type = type(
            "PKObject",
            (Object,),
            {
                "pk1": ObjectField.Checked("pk_1", check(int), pk=0),
                "pk2": ObjectField.Checked("pk_2", check(int), pk=1),
            },
        )
        error = self.assertRaises(TypeError, object_type, 0)
        self.assertThat(
            str(error), Equals("data must be a mapping or a sequence, not int")
        )

    def test__init_validates_property_when_multiple_pks_mapping(self):
        object_type = type(
            "PKObject",
            (Object,),
            {
                "pk1": ObjectField.Checked("pk_1", check(int), pk=0),
                "pk2": ObjectField.Checked("pk_2", check(int), pk=1),
            },
        )
        error = self.assertRaises(
            TypeError, object_type, {"pk_1": 0, "pk_2": "bad", "__incomplete__": True}
        )
        self.assertThat(str(error), Equals("'bad' is not of type %r" % int))

    def test__init_validates_property_when_multiple_alt_pks_mapping(self):
        object_type = type(
            "PKObject",
            (Object,),
            {
                "pk1": ObjectField.Checked("pk_1", check(int), pk=0),
                "pk2": ObjectField.Checked("pk_2", check(int), pk=1),
                "alt_pk1": ObjectField.Checked("alt_pk_1", check(int), alt_pk=0),
                "alt_pk2": ObjectField.Checked("alt_pk_2", check(int), alt_pk=1),
            },
        )
        error = self.assertRaises(
            TypeError,
            object_type,
            {"alt_pk_1": 0, "alt_pk_2": "bad", "__incomplete__": True},
        )
        self.assertThat(str(error), Equals("'bad' is not of type %r" % int))

    def test__init_validates_property_when_multiple_pks_sequence(self):
        object_type = type(
            "PKObject",
            (Object,),
            {
                "pk1": ObjectField.Checked("pk_1", check(int), pk=0),
                "pk2": ObjectField.Checked("pk_2", check(int), pk=1),
            },
        )
        error = self.assertRaises(TypeError, object_type, [0, "bad"])
        self.assertThat(str(error), Equals("'bad' is not of type %r" % int))

    def test__loaded(self):
        object_a = Object({})
        self.assertTrue(object_a.loaded)
        object_a._loaded = False
        self.assertFalse(object_a.loaded)

    def test__cannot_access_attributes_when_unloaded(self):
        object_type = type(
            "PKObject",
            (Object,),
            {"pk": ObjectField("pk_d", pk=True), "name": ObjectField("name")},
        )
        object_pk = randint(0, 20)
        object_a = object_type(object_pk)
        object_a._data["name"] = make_name("name")
        self.assertFalse(object_a.loaded)
        error = self.assertRaises(ObjectNotLoaded, getattr, object_a, "name")
        self.assertThat(
            str(error), Equals("cannot access attribute 'name' of object 'PKObject'")
        )

    def test__can_access_pk_attributes_when_unloaded(self):
        object_type = type(
            "PKObject",
            (Object,),
            {"pk": ObjectField("pk_d", pk=True), "name": ObjectField("name")},
        )
        object_pk = randint(0, 20)
        object_a = object_type(object_pk)
        self.assertFalse(object_a.loaded)
        self.assertEquals(object_pk, object_a.pk)

    def test__can_access_alt_pk_attributes_when_unloaded(self):
        object_type = type(
            "PKObject",
            (Object,),
            {
                "pk": ObjectField("pk_d", pk=True),
                "alt_pk": ObjectField("alt_pk_d", alt_pk=True),
                "name": ObjectField("name"),
            },
        )
        object_alt_pk = randint(0, 20)
        object_a = object_type({"alt_pk_d": object_alt_pk, "__incomplete__": True})
        self.assertFalse(object_a.loaded)
        self.assertEquals(object_alt_pk, object_a.alt_pk)

    def test__can_access_attributes_when_loaded(self):
        object_type = type(
            "PKObject",
            (Object,),
            {"pk": ObjectField("pk_d", pk=True), "name": ObjectField("name")},
        )
        object_pk = randint(0, 20)
        object_name = make_name("name")
        object_a = object_type({"pk_d": object_pk, "name": object_name})
        self.assertTrue(object_a.loaded)
        self.assertEquals(object_name, object_a.name)

    def test__equal_when_data_matches(self):
        data = {"key": make_name("value")}
        object_a = Object(data)
        object_b = Object(data)
        self.assertThat(object_a, Equals(object_b))
        self.assertThat(object_b, Equals(object_a))

    def test__not_equal_when_types_different(self):
        # Even if one is a subclass of the other.
        data = {"key": make_name("value")}
        object_a = Object(data)
        object_b = type("Object", (Object,), {})(data)
        self.assertThat(object_a, Not(Equals(object_b)))
        self.assertThat(object_b, Not(Equals(object_a)))

    def test__string_representation_includes_field_values(self):
        class Example(Object):
            alice = ObjectField("alice")
            bob = ObjectField("bob")

        example = Example(
            {
                "alice": make_name_without_spaces("alice"),
                "bob": make_name_without_spaces("bob"),
            }
        )

        self.assertThat(
            repr(example),
            Equals("<Example alice=%(alice)r bob=%(bob)r>" % example._data),
        )

    def test__string_representation_can_be_limited_to_selected_fields(self):
        class Example(Object):
            alice = ObjectField("alice")
            bob = ObjectField("bob")

        example = Example(
            {
                "alice": make_name_without_spaces("alice"),
                "bob": make_name_without_spaces("bob"),
            }
        )

        # A string repr can be prepared using only the "alice" field.
        self.assertThat(
            example.__repr__(fields={"alice"}),
            Equals("<Example alice=%(alice)r>" % example._data),
        )

        # Fields are always displayed in a stable order though.
        self.assertThat(
            example.__repr__(fields=["bob", "alice"]),
            Equals("<Example alice=%(alice)r bob=%(bob)r>" % example._data),
        )

    def test_refresh_raises_AttributeError_when_no_read_defined(self):
        object_a = Object({})
        error = self.assertRaises(AttributeError, object_a.refresh)
        self.assertThat(str(error), Equals("'Object' object doesn't support refresh."))

    def test_refresh_with_one_pk(self):
        object_type = type("PKObject", (Object,), {"pk": ObjectField("pk_d", pk=True)})
        object_pk = randint(0, 20)
        new_data = {"pk_d": object_pk, "other": randint(0, 20)}
        mock_read = AsyncCallableMock(return_value=object_type(new_data))
        self.patch(object_type, "read", mock_read)
        object_a = object_type(object_pk)
        self.assertFalse(object_a.loaded)
        object_a.refresh()
        self.assertTrue(object_a.loaded)
        self.assertThat(object_a._data, Equals(new_data))
        self.assertThat(object_a._orig_data, Equals(new_data))
        self.assertThat(object_a._orig_data, Not(Is(object_a._data)))
        self.assertThat(object_a._changed_data, Equals({}))
        self.assertThat(mock_read.call_args_list, Equals([call(object_pk)]))

    def test_refresh_with_one_alt_pk(self):
        object_type = type(
            "PKObject",
            (Object,),
            {
                "pk": ObjectField("pk_d", pk=True),
                "alt_pk": ObjectField("alt_pk_d", alt_pk=True),
            },
        )
        object_pk = randint(0, 20)
        object_alt_pk = randint(0, 20)
        new_data = {
            "pk_d": object_pk,
            "alt_pk_d": object_alt_pk,
            "other": randint(0, 20),
        }
        mock_read = AsyncCallableMock(return_value=object_type(new_data))
        self.patch(object_type, "read", mock_read)
        object_a = object_type({"alt_pk_d": object_alt_pk, "__incomplete__": True})
        self.assertFalse(object_a.loaded)
        object_a.refresh()
        self.assertTrue(object_a.loaded)
        self.assertThat(object_a._data, Equals(new_data))
        self.assertThat(object_a._orig_data, Equals(new_data))
        self.assertThat(object_a._orig_data, Not(Is(object_a._data)))
        self.assertThat(object_a._changed_data, Equals({}))
        self.assertThat(mock_read.call_args_list, Equals([call(object_alt_pk)]))

    def test_refresh_with_multiple_pk(self):
        object_type = type(
            "PKObject",
            (Object,),
            {"pk1": ObjectField("pk_1", pk=0), "pk2": ObjectField("pk_2", pk=1)},
        )
        object_pk_1 = randint(0, 20)
        object_pk_2 = randint(0, 20)
        new_data = {"pk_1": object_pk_1, "pk_2": object_pk_2, "other": randint(0, 20)}
        mock_read = AsyncCallableMock(return_value=object_type(new_data))
        self.patch(object_type, "read", mock_read)
        object_a = object_type([object_pk_1, object_pk_2])
        self.assertFalse(object_a.loaded)
        object_a.refresh()
        self.assertTrue(object_a.loaded)
        self.assertThat(object_a._data, Equals(new_data))
        self.assertThat(object_a._orig_data, Equals(new_data))
        self.assertThat(object_a._orig_data, Not(Is(new_data)))
        self.assertThat(object_a._changed_data, Equals({}))
        self.assertThat(
            mock_read.call_args_list, Equals([call(object_pk_1, object_pk_2)])
        )

    def test_refresh_with_multiple_alt_pk(self):
        object_type = type(
            "PKObject",
            (Object,),
            {
                "pk1": ObjectField("pk_1", pk=0),
                "pk2": ObjectField("pk_2", pk=1),
                "alt_pk1": ObjectField("alt_pk_1", alt_pk=0),
                "alt_pk2": ObjectField("alt_pk_2", alt_pk=1),
            },
        )
        object_pk_1 = randint(0, 20)
        object_pk_2 = randint(0, 20)
        object_alt_pk_1 = randint(0, 20)
        object_alt_pk_2 = randint(0, 20)
        new_data = {
            "pk_1": object_pk_1,
            "pk_2": object_pk_2,
            "alt_pk_1": object_alt_pk_1,
            "alt_pk_2": object_alt_pk_2,
            "other": randint(0, 20),
        }
        mock_read = AsyncCallableMock(return_value=object_type(new_data))
        self.patch(object_type, "read", mock_read)
        object_a = object_type(
            {
                "alt_pk_1": object_alt_pk_1,
                "alt_pk_2": object_alt_pk_2,
                "__incomplete__": True,
            }
        )
        self.assertFalse(object_a.loaded)
        object_a.refresh()
        self.assertTrue(object_a.loaded)
        self.assertThat(object_a._data, Equals(new_data))
        self.assertThat(object_a._orig_data, Equals(new_data))
        self.assertThat(object_a._orig_data, Not(Is(new_data)))
        self.assertThat(object_a._changed_data, Equals({}))
        self.assertThat(
            mock_read.call_args_list, Equals([call(object_alt_pk_1, object_alt_pk_2)])
        )

    def test_refresh_raises_AttributeError_when_no_pk_fields(self):
        object_type = type("PKObject", (Object,), {})
        mock_read = AsyncCallableMock(return_value=object_type({}))
        self.patch(object_type, "read", mock_read)
        object_a = object_type({})
        error = self.assertRaises(AttributeError, object_a.refresh)
        self.assertThat(
            str(error),
            Equals("unable to perform 'refresh' no primary key fields defined."),
        )

    def test_refresh_raises_TypeError_on_mismatch(self):
        object_type = type("PKObject", (Object,), {"pk": ObjectField("pk_d", pk=True)})
        mock_read = AsyncCallableMock(return_value=Object({}))
        self.patch(object_type, "read", mock_read)
        object_a = object_type({"pk_d": 0})
        error = self.assertRaises(TypeError, object_a.refresh)
        self.assertThat(
            str(error),
            Equals("result of 'PKObject.read' must be 'PKObject', not 'Object'"),
        )

    def test_save_raises_AttributeError_when_handler_has_no_update(self):
        object_type = type("NotSaveableObject", (Object,), {"_handler": object()})
        error = self.assertRaises(AttributeError, object_type({}).save)
        self.assertThat(
            str(error), Equals("'NotSaveableObject' object doesn't support save.")
        )

    def test_save_does_nothing_when_nothing_changed(self):
        handler = Mock()
        handler.update = AsyncCallableMock(return_value={})
        object_type = type(
            "SaveableObject",
            (Object,),
            {"_handler": handler, "name": ObjectField("name")},
        )
        object_a = object_type({})
        object_a.save()
        self.assertThat(handler.update.call_count, Equals(0))

    def test_save_calls_update_on_handler_with_params(self):
        object_id = randint(0, 10)
        saved_name = make_name("name")
        updated_data = {"id": object_id, "name": saved_name}
        handler = Mock()
        handler.params = ["id"]
        handler.update = AsyncCallableMock(return_value=updated_data)
        object_type = type(
            "SaveableObject",
            (Object,),
            {"_handler": handler, "name": ObjectField("name")},
        )
        object_a = object_type({"id": object_id})
        new_name = make_name("new")
        object_a.name = new_name
        object_a.save()
        self.assertThat(
            handler.update.call_args_list, Equals([call(id=object_id, name=new_name)])
        )
        self.assertThat(object_a._data, Equals(updated_data))
        self.assertThat(object_a._orig_data, Equals(updated_data))
        self.assertThat(object_a._orig_data, Not(Is(object_a._data)))
        self.assertThat(object_a._changed_data, Equals({}))


class TestObjectSet(TestCase):
    """Tests for `ObjectSet`."""

    def test__defines_slots(self):
        self.assertThat(ObjectSet.__slots__, Equals(("__weakref__", "_items")))

    def test__inherits_ObjectBasics(self):
        self.assertThat(ObjectSet.__mro__, Contains(ObjectBasics))

    def test__init_sets__items_from_sequence(self):
        items = [{"alice": make_name_without_spaces("alice")}]
        self.assertThat(ObjectSet(items)._items, Equals(items))

    def test__init_sets__items_from_iterable(self):
        items = [{"alice": make_name_without_spaces("alice")}]
        self.assertThat(ObjectSet(iter(items))._items, Equals(items))

    def test__init_rejects_mapping(self):
        error = self.assertRaises(TypeError, ObjectSet, {})
        self.assertThat(str(error), Equals("data must be sequence-like, not dict"))

    def test__init_rejects_str(self):
        error = self.assertRaises(TypeError, ObjectSet, "")
        self.assertThat(str(error), Equals("data must be sequence-like, not str"))

    def test__init_rejects_bytes(self):
        error = self.assertRaises(TypeError, ObjectSet, b"")
        self.assertThat(str(error), Equals("data must be sequence-like, not bytes"))

    def test__init_rejects_non_iterable(self):
        error = self.assertRaises(TypeError, ObjectSet, 123)
        self.assertThat(str(error), Equals("data must be sequence-like, not int"))

    def test__length_is_number_of_items(self):
        items = [0] * randrange(0, 100)
        objectset = ObjectSet(items)
        self.assertThat(objectset, HasLength(len(items)))

    def test__can_be_indexed(self):
        items = [make_name_without_spaces(str(index)) for index in range(5)]
        objectset = ObjectSet(items)
        for index, item in enumerate(items):
            self.assertThat(objectset[index], Equals(item))

    def test__can_be_sliced(self):
        items = [make_name_without_spaces(str(index)) for index in range(5)]
        objectset1 = ObjectSet(items)
        objectset2 = objectset1[1:3]
        self.assertThat(objectset2, IsInstance(ObjectSet))
        self.assertThat(list(objectset2), Equals(items[1:3]))

    def test__iteration_yield_items(self):
        items = [make_name_without_spaces(str(index)) for index in range(5)]
        objectset = ObjectSet(items)
        self.assertThat(list(objectset), Equals(items))

    def test__reversed_yields_items_in_reverse(self):
        items = [make_name_without_spaces(str(index)) for index in range(5)]
        objectset = ObjectSet(items)
        self.assertThat(list(reversed(objectset)), Equals(items[::-1]))

    def test__membership_can_be_tested(self):
        item1 = make_name_without_spaces("item")
        item2 = make_name_without_spaces("item")
        objectset = ObjectSet([item1])
        self.assertThat(objectset, Contains(item1))
        self.assertThat(objectset, Not(Contains(item2)))

    def test__equal_when_items_match(self):
        items = [{"key": make_name("value")}]
        objectset_a = ObjectSet(items)
        objectset_b = ObjectSet(items)
        self.assertThat(objectset_a, Equals(objectset_b))
        self.assertThat(objectset_b, Equals(objectset_a))

    def test__not_equal_when_types_different(self):
        # Even if one is a subclass of the other.
        items = [{"key": make_name("value")}]
        objectset_a = ObjectSet(items)
        objectset_b = type("ObjectSet", (ObjectSet,), {})(items)
        self.assertThat(objectset_a, Not(Equals(objectset_b)))
        self.assertThat(objectset_b, Not(Equals(objectset_a)))

    def test__string_representation_includes_length_and_items(self):
        class Example(Object):
            alice = ObjectField("alice")

        class ExampleSet(ObjectSet):
            pass

        example = ExampleSet(
            [Example({"alice": "wonderland"}), Example({"alice": "cooper"})]
        )

        self.assertThat(
            repr(example),
            Equals(
                "<ExampleSet length=2 items=["
                "<Example alice='wonderland'>, <Example alice='cooper'>"
                "]>"
            ),
        )


class TestObjectField(TestCase):
    """Tests for `ObjectField`."""

    def test__gets_sets_and_deletes_the_given_name_from_object(self):
        class Example(Object):
            alice = ObjectField("alice")

        example = Example({})

        # At first, referencing "alice" yields an exception.
        self.assertRaises(AttributeError, getattr, example, "alice")
        self.assertThat(example._data, Equals({}))

        # Setting "alice" stores the value in the object's _data dict.
        example.alice = sentinel.alice
        self.assertThat(example.alice, Is(sentinel.alice))
        self.assertThat(example._data, Equals({"alice": sentinel.alice}))

        # Deleting "alice" removes the value from the object's _data dict, and
        # referencing "alice" yields an exception.
        del example.alice
        self.assertRaises(AttributeError, getattr, example, "alice")
        self.assertThat(example._data, Equals({}))

    def test__default_is_returned_when_value_not_found_in_object(self):
        class Example(Object):
            alice = ObjectField("alice", default=sentinel.alice_default)

        example = Example({})

        # At first, referencing "alice" yields the default value.
        self.assertThat(example.alice, Is(sentinel.alice_default))
        self.assertThat(example._data, Equals({}))

        # Setting "alice" stores the value in the object's _data dict.
        example.alice = sentinel.alice
        self.assertThat(example, MatchesStructure(alice=Is(sentinel.alice)))
        self.assertThat(example._data, Equals({"alice": sentinel.alice}))

        # Deleting "alice" removes the value from the object's _data dict, and
        # referencing "alice" again yields the default value.
        del example.alice
        self.assertThat(example.alice, Is(sentinel.alice_default))
        self.assertThat(example._data, Equals({}))

    def test__readonly_prevents_setting_or_deleting(self):
        class Example(Object):
            alice = ObjectField("alice", readonly=True)

        example = Example({"alice": sentinel.in_wonderland})

        self.assertThat(example.alice, Is(sentinel.in_wonderland))
        self.assertRaises(AttributeError, setattr, example, "alice", 123)
        self.assertThat(example.alice, Is(sentinel.in_wonderland))
        self.assertRaises(AttributeError, delattr, example, "alice")
        self.assertThat(example.alice, Is(sentinel.in_wonderland))

    def test__conversion_and_validation_happens(self):
        class AliceField(ObjectField):
            """A most peculiar field."""

            def datum_to_value(self, instance, datum):
                return datum + instance.datum_to_value_delta

            def value_to_datum(self, instance, value):
                return value + instance.value_to_datum_delta

        class Example(Object):
            alice = AliceField("alice")
            # Deltas to apply to datums and values.
            datum_to_value_delta = 2
            value_to_datum_delta = 3

        example = Example({})
        example.alice = 0

        # value_to_datum_delta was added to the value we specified before
        # being stored in the object's _data dict.
        self.assertThat(example._data, Equals({"alice": 3}))

        # datum_to_value_delta is added to the datum in the object's _data
        # dict before being returned to us.
        self.assertThat(example.alice, Equals(5))

    def test__default_is_not_subject_to_conversion_or_validation(self):
        class AliceField(ObjectField):
            """Another most peculiar field."""

        class Example(Object):
            alice = AliceField("alice", default=sentinel.alice_default)

        example = Example({})

        # The default given is treated as a Python-side value rather than a
        # MAAS-side datum, so is not passed through datum_to_value (or
        # value_to_datum for that matter).
        self.assertThat(example.alice, Is(sentinel.alice_default))

    def test__set_new_value_is_set_in_changed(self):
        class AliceField(ObjectField):
            """Another most peculiar field."""

        class Example(Object):
            alice = AliceField("alice")

        example = Example({})

        example.alice = sentinel.alice
        self.assertThat(example._changed_data, Equals({"alice": sentinel.alice}))

    def test__set_update_value_is_set_in_changed(self):
        class AliceField(ObjectField):
            """Another most peculiar field."""

        class Example(Object):
            alice = AliceField("alice")

        example = Example({"alice": sentinel.alice})

        example.alice = sentinel.new_alice
        self.assertThat(example._changed_data, Equals({"alice": sentinel.new_alice}))

    def test__set_update_value_replaces_changed(self):
        class AliceField(ObjectField):
            """Another most peculiar field."""

        class Example(Object):
            alice = AliceField("alice")

        example = Example({"alice": sentinel.alice})

        example.alice = sentinel.new_alice
        self.assertThat(example._changed_data, Equals({"alice": sentinel.new_alice}))
        example.alice = sentinel.newer_alice
        self.assertThat(example._changed_data, Equals({"alice": sentinel.newer_alice}))

    def test__set_update_value_to_orig_removes_from_changed(self):
        class AliceField(ObjectField):
            """Another most peculiar field."""

        class Example(Object):
            alice = AliceField("alice")

        alice = make_name_without_spaces("alice")
        new_alice = make_name_without_spaces("alice")
        example = Example({"alice": alice})

        example.alice = new_alice
        self.assertThat(example._changed_data, Equals({"alice": new_alice}))
        example.alice = alice
        self.assertThat(example._changed_data, Equals({}))

    def test__delete_marks_field_deleted(self):
        class AliceField(ObjectField):
            """Another most peculiar field."""

        class Example(Object):
            alice = AliceField("alice")

        example = Example({"alice": sentinel.alice})

        del example.alice
        self.assertThat(example._changed_data, Equals({"alice": None}))

    def test__set_deleted_field_sets_changed(self):
        class AliceField(ObjectField):
            """Another most peculiar field."""

        class Example(Object):
            alice = AliceField("alice")

        alice = make_name_without_spaces("alice")
        new_alice = make_name_without_spaces("alice")
        example = Example({"alice": alice})

        del example.alice
        example.alice = new_alice
        self.assertThat(example._changed_data, Equals({"alice": new_alice}))
        example.alice = alice
        self.assertThat(example._changed_data, Equals({}))


class TestObjectFieldChecked(TestCase):
    """Tests for `ObjectField.Checked`."""

    def test__creates_subclass(self):
        field = ObjectField.Checked("alice")
        self.assertThat(type(field).__mro__, Contains(ObjectField))
        self.assertThat(type(field), Not(Is(ObjectField)))
        self.assertThat(type(field).__name__, Equals("ObjectField.Checked#alice"))

    def test__overrides_datum_to_value(self):
        add_one = lambda value: value + 1
        field = ObjectField.Checked("alice", datum_to_value=add_one)
        self.assertThat(field.datum_to_value(None, 1), Equals(2))

    def test__overrides_value_to_daturm(self):
        add_one = lambda value: value + 1
        field = ObjectField.Checked("alice", value_to_datum=add_one)
        self.assertThat(field.value_to_datum(None, 1), Equals(2))

    def test__works_in_place(self):

        # Deltas to apply to datums and values.
        datum_to_value_delta = 2
        value_to_datum_delta = 3

        class Example(Object):
            alice = ObjectField.Checked(
                "alice",
                (lambda datum: datum + datum_to_value_delta),
                (lambda value: value + value_to_datum_delta),
            )

        example = Example({})
        example.alice = 0

        # value_to_datum_delta was added to the value we specified before
        # being stored in the object's _data dict.
        self.assertThat(example._data, Equals({"alice": 3}))

        # datum_to_value_delta is added to the datum in the object's _data
        # dict before being returned to us.
        self.assertThat(example.alice, Equals(5))


class TestObjectFieldRelated(TestCase):
    """Tests for `ObjectFieldRelated`."""

    def test__init__requires_str_or_Object_class(self):
        self.assertRaises(TypeError, ObjectFieldRelated, "name", 0)
        self.assertRaises(TypeError, ObjectFieldRelated, "name", object)
        # Doesn't raise error.
        ObjectFieldRelated("name", "class")
        ObjectFieldRelated("name", Object)

    def test_datum_to_value_returns_None_on_None(self):
        self.assertIsNone(
            ObjectFieldRelated("name", "class").datum_to_value(object(), None)
        )

    def test_datum_to_value_converts_to_bound_class(self):
        rel_object_type = type(
            "RelObject", (Object,), {"pk": ObjectField("pk_d", pk=True)}
        )
        origin = Mock()
        origin.RelObject = rel_object_type
        instance = type("InstObject", (Object,), {"_origin": origin})({})
        field = ObjectFieldRelated("related_id", "RelObject")
        rel_id = randint(0, 20)
        rel_object = field.datum_to_value(instance, rel_id)
        self.assertIsInstance(rel_object, rel_object_type)
        self.assertFalse(rel_object.loaded)
        self.assertThat(
            rel_object._data, Equals({"instobject": instance, "pk_d": rel_id})
        )

    def test_datum_to_value_uses_reverse_name(self):
        rel_object_type = type(
            "RelObject", (Object,), {"pk": ObjectField("pk_d", pk=True)}
        )
        origin = Mock()
        origin.RelObject = rel_object_type
        instance = type("Object", (Object,), {"_origin": origin})({})
        field = ObjectFieldRelated("related_id", "RelObject", reverse="reverse")
        rel_id = randint(0, 20)
        rel_object = field.datum_to_value(instance, rel_id)
        self.assertIsInstance(rel_object, rel_object_type)
        self.assertFalse(rel_object.loaded)
        self.assertThat(rel_object._data, Equals({"reverse": instance, "pk_d": rel_id}))

    def test_datum_to_value_doesnt_include_reverse(self):
        rel_object_type = type(
            "RelObject", (Object,), {"pk": ObjectField("pk_d", pk=True)}
        )
        origin = Mock()
        origin.RelObject = rel_object_type
        instance = type("Object", (Object,), {"_origin": origin})({})
        field = ObjectFieldRelated("related_id", "RelObject", reverse=None)
        rel_id = randint(0, 20)
        rel_object = field.datum_to_value(instance, rel_id)
        self.assertIsInstance(rel_object, rel_object_type)
        self.assertFalse(rel_object.loaded)
        self.assertThat(rel_object._data, Equals({"pk_d": rel_id}))

    def test_value_to_datum_returns_None_on_None(self):
        self.assertIsNone(
            ObjectFieldRelated("name", "class").value_to_datum(object(), None)
        )

    def test_value_to_datum_requires_same_class(self):
        rel_object_type = type(
            "RelObject", (Object,), {"pk": ObjectField("pk_d", pk=True)}
        )
        origin = Mock()
        origin.RelObject = rel_object_type
        instance = type("Object", (Object,), {"_origin": origin})({})
        field = ObjectFieldRelated("related_id", "RelObject")
        error = self.assertRaises(TypeError, field.value_to_datum, instance, object())
        self.assertThat(str(error), Equals("must be RelObject, not object"))

    def test_value_to_datum_with_one_primary_key(self):
        rel_object_type = type(
            "RelObject", (Object,), {"pk": ObjectField("pk_d", pk=True)}
        )
        origin = Mock()
        origin.RelObject = rel_object_type
        instance = type("Object", (Object,), {"_origin": origin})({})
        field = ObjectFieldRelated("related_id", "RelObject")
        rel_id = randint(0, 20)
        rel_object = rel_object_type(rel_id)
        self.assertThat(field.value_to_datum(instance, rel_object), Equals(rel_id))

    def test_value_to_datum_with_multiple_primary_keys(self):
        rel_object_type = type(
            "RelObject",
            (Object,),
            {"pk1": ObjectField("pk_1", pk=0), "pk2": ObjectField("pk_2", pk=1)},
        )
        origin = Mock()
        origin.RelObject = rel_object_type
        instance = type("Object", (Object,), {"_origin": origin})({})
        field = ObjectFieldRelated("related", "RelObject")
        rel_1 = randint(0, 20)
        rel_2 = randint(0, 20)
        rel_object = rel_object_type((rel_1, rel_2))
        self.assertThat(
            field.value_to_datum(instance, rel_object), Equals((rel_1, rel_2))
        )

    def test_value_to_datum_raises_error_when_no_primary_keys(self):
        rel_object_type = type("RelObject", (Object,), {"pk": ObjectField("pk_d")})
        origin = Mock()
        origin.RelObject = rel_object_type
        instance = type("Object", (Object,), {"_origin": origin})({})
        field = ObjectFieldRelated("related_id", "RelObject")
        rel_id = randint(0, 20)
        rel_object = rel_object_type({"pk_d": rel_id})
        error = self.assertRaises(
            AttributeError, field.value_to_datum, instance, rel_object
        )
        self.assertThat(
            str(error),
            Equals(
                "unable to perform set object no primary key "
                "fields defined for RelObject"
            ),
        )


class TestObjectFieldRelatedSet(TestCase):
    """Tests for `ObjectFieldRelatedSet`."""

    def test__init__requires_str_or_Object_class(self):
        self.assertRaises(TypeError, ObjectFieldRelatedSet, "name", 0)
        self.assertRaises(TypeError, ObjectFieldRelatedSet, "name", object)
        # Doesn't raise error.
        ObjectFieldRelatedSet("name", "class")
        ObjectFieldRelatedSet("name", ObjectSet)

    def test_datum_to_value_returns_empty_list_on_None(self):
        self.assertEquals(
            ObjectFieldRelatedSet("name", "class").datum_to_value(object(), None), []
        )

    def test_datum_must_be_a_sequence(self):
        field = ObjectFieldRelatedSet("name", "class")
        error = self.assertRaises(TypeError, field.datum_to_value, object(), 0)
        self.assertThat(str(error), Equals("datum must be a sequence, not int"))

    def test_datum_to_value_converts_to_set_of_bound_class(self):
        rel_object_type = type(
            "RelObject", (Object,), {"pk": ObjectField("pk_d", pk=True)}
        )
        rel_object_set_type = type(
            "RelObjectSet", (ObjectSet,), {"_object": rel_object_type}
        )
        origin = Mock()
        origin.RelObjectSet = rel_object_set_type
        instance = type("InstObject", (Object,), {"_origin": origin})({})
        field = ObjectFieldRelatedSet("related_ids", "RelObjectSet")
        rel_ids = range(5)
        rel_object_set = field.datum_to_value(instance, rel_ids)
        self.assertEquals(5, len(rel_object_set))
        self.assertIsInstance(rel_object_set[0], rel_object_type)
        self.assertFalse(rel_object_set[0].loaded)
        self.assertThat(
            rel_object_set[0]._data,
            Equals({"instobject": instance, "pk_d": rel_ids[0]}),
        )

    def test_datum_to_value_uses_reverse_name(self):
        rel_object_type = type(
            "RelObject", (Object,), {"pk": ObjectField("pk_d", pk=True)}
        )
        rel_object_set_type = type(
            "RelObjectSet", (ObjectSet,), {"_object": rel_object_type}
        )
        origin = Mock()
        origin.RelObjectSet = rel_object_set_type
        instance = type("InstObject", (Object,), {"_origin": origin})({})
        field = ObjectFieldRelatedSet("related_ids", "RelObjectSet", reverse="reverse")
        rel_ids = range(5)
        rel_object_set = field.datum_to_value(instance, rel_ids)
        self.assertEquals(5, len(rel_object_set))
        self.assertIsInstance(rel_object_set[0], rel_object_type)
        self.assertFalse(rel_object_set[0].loaded)
        self.assertThat(
            rel_object_set[0]._data, Equals({"reverse": instance, "pk_d": rel_ids[0]})
        )

    def test_datum_to_value_doesnt_include_reverse(self):
        rel_object_type = type(
            "RelObject", (Object,), {"pk": ObjectField("pk_d", pk=True)}
        )
        rel_object_set_type = type(
            "RelObjectSet", (ObjectSet,), {"_object": rel_object_type}
        )
        origin = Mock()
        origin.RelObjectSet = rel_object_set_type
        instance = type("InstObject", (Object,), {"_origin": origin})({})
        field = ObjectFieldRelatedSet("related_ids", "RelObjectSet", reverse=None)
        rel_ids = range(5)
        rel_object_set = field.datum_to_value(instance, rel_ids)
        self.assertEquals(5, len(rel_object_set))
        self.assertIsInstance(rel_object_set[0], rel_object_type)
        self.assertFalse(rel_object_set[0].loaded)
        self.assertThat(rel_object_set[0]._data, Equals({"pk_d": rel_ids[0]}))

    def test_datum_to_value_wraps_managed_create(self):
        rel_object_type = type(
            "RelObject", (Object,), {"pk": ObjectField("pk_d", pk=True)}
        )
        create_mock = AsyncCallableMock(return_value=rel_object_type({}))
        rel_object_set_type = type(
            "RelObjectSet",
            (ObjectSet,),
            {"_object": rel_object_type, "create": create_mock},
        )
        origin = Mock()
        origin.RelObjectSet = rel_object_set_type
        instance = type("InstObject", (Object,), {"_origin": origin})(
            {"related_ids": []}
        )
        field = ObjectFieldRelatedSet("related_ids", "RelObjectSet")
        rel_ids = range(5)
        rel_object_set = field.datum_to_value(instance, rel_ids)
        rel_object_set.create()
        self.assertEquals(
            "RelObjectSet.Managed#InstObject", type(rel_object_set).__name__
        )
        self.assertThat(create_mock.call_args_list, Equals([call(instance)]))


class TestOriginBase(TestCase):
    """Tests for `OriginBase`."""

    def test__session_is_underlying_session(self):
        profile = make_profile()
        session = bones.SessionAPI.fromProfile(profile)
        origin = OriginBase(session)
        self.assertThat(origin.session, Is(session))

    def test__session_is_read_only(self):
        profile = make_profile()
        session = bones.SessionAPI.fromProfile(profile)
        origin = OriginBase(session)
        self.assertRaises(AttributeError, setattr, origin, "session", None)



# Copyright 2016-2017 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for `maas.client.utils.profiles`."""

import contextlib
from pathlib import Path
import sqlite3

from testtools.matchers import Equals, Is, Not
from twisted.python.filepath import FilePath

from ...testing import make_name_without_spaces, TestCase
from ..profiles import Profile, ProfileNotFound, ProfileStore
from ..testing import make_Credentials


def make_profile(name=None):
    if name is None:
        name = make_name_without_spaces("name")
    return Profile(
        name=name,
        url="http://example.com:5240/",
        credentials=make_Credentials(),
        description={"resources": []},
        something=make_name_without_spaces("something"),
    )


class TestProfile(TestCase):
    def test__instances_are_immutable(self):
        profile = make_profile()
        self.assertRaises(AttributeError, setattr, profile, "name", "foo")
        self.assertRaises(AttributeError, setattr, profile, "uri", "foo")
        self.assertRaises(AttributeError, setattr, profile, "bar", "foo")

    def test__replace_returns_a_new_profile(self):
        profile1 = make_profile()
        profile2 = profile1.replace()
        self.assertThat(profile2, Not(Is(profile1)))
        self.assertThat(profile2.name, Equals(profile1.name))
        self.assertThat(profile2.url, Equals(profile1.url))
        self.assertThat(profile2.credentials, Equals(profile1.credentials))
        self.assertThat(profile2.description, Equals(profile1.description))
        self.assertThat(profile2.other, Equals(profile1.other))

    def test__replace_returns_a_new_profile_with_modifications(self):
        profile1 = make_profile()
        profile2 = profile1.replace(name=profile1.name + "basil", hello="world")
        self.assertThat(profile2.name, Equals(profile1.name + "basil"))
        self.assertThat(profile2.other, Equals(dict(profile1.other, hello="world")))

    def test__dump_returns_dict_with_all_state(self):
        profile = make_profile()
        self.assertThat(
            profile.dump(),
            Equals(
                {
                    "name": profile.name,
                    "url": profile.url,
                    "credentials": profile.credentials,
                    "description": profile.description,
                    "something": profile.other["something"],
                }
            ),
        )

    def test__representation(self):
        profile = make_profile()
        self.assertThat(
            repr(profile), Equals("<Profile {0.name} {0.url}>".format(profile))
        )

    def test__representation_of_anonymous_profile(self):
        profile = make_profile().replace(credentials=None)
        self.assertThat(
            repr(profile),
            Equals("<Profile {0.name} (anonymous) {0.url}>".format(profile)),
        )


class TestProfileStore(TestCase):
    """Tests for `ProfileStore`."""

    def setUp(self):
        super(TestProfileStore, self).setUp()
        self.null_profile = Path("/dev/null")

    def test_init(self):
        database = sqlite3.connect(":memory:")
        config = ProfileStore(database)
        # The profiles table has been created.
        self.assertEqual(
            config.database.execute(
                "SELECT COUNT(*) FROM sqlite_master"
                " WHERE type = 'table'"
                "   AND name = 'profiles'"
            ).fetchone(),
            (1,),
        )

    def test_profiles_pristine(self):
        # A pristine configuration has no profiles.
        database = sqlite3.connect(":memory:")
        config = ProfileStore(database)
        self.assertSetEqual(set(), set(config))

    def test_saving_and_loading_profile(self):
        database = sqlite3.connect(":memory:")
        config = ProfileStore(database)
        profile = make_profile()
        config.save(profile)
        self.assertEqual({profile.name}, set(config))
        self.assertEqual(profile, config.load(profile.name))

    def test_saving_failure_does_not_corrupt_database(self):
        database = sqlite3.connect(":memory:")
        config = ProfileStore(database)
        profile_good = make_profile()
        config.save(profile_good)
        profile_bad = profile_good.replace(foo=object())
        self.assertRaises(TypeError, config.save, profile_bad)
        self.assertEqual({profile_good.name}, set(config))
        self.assertEqual(profile_good, config.load(profile_good.name))

    def test_replacing_profile(self):
        database = sqlite3.connect(":memory:")
        config = ProfileStore(database)
        profile1 = make_profile().replace(name="alice")
        profile2 = make_profile().replace(name="alice")
        self.assertNotEqual(profile1, profile2)
        config.save(profile1)
        config.save(profile2)
        self.assertEqual({"alice"}, set(config))
        self.assertEqual(profile2, config.load("alice"))

    def test_loading_non_existent_profile(self):
        database = sqlite3.connect(":memory:")
        config = ProfileStore(database)
        self.assertRaises(ProfileNotFound, config.load, "alice")

    def test_removing_profile(self):
        database = sqlite3.connect(":memory:")
        config = ProfileStore(database)
        profile = make_profile()
        config.save(profile)
        config.delete(profile.name)
        self.assertEqual(set(), set(config))

    def test_open_and_close(self):
        # ProfileStore.open() returns a context manager that closes the
        # database on exit.
        config_file = self.makeDir().joinpath("config")
        config = ProfileStore.open(config_file)
        self.assertIsInstance(config, contextlib._GeneratorContextManager)
        with config as config:
            self.assertIsInstance(config, ProfileStore)
            self.assertEqual((1,), config.database.execute("SELECT 1").fetchone())
        self.assertRaises(sqlite3.ProgrammingError, config.database.execute, "SELECT 1")

    def test_open_permissions_new_database(self):
        # ProfileStore.open() applies restrictive file permissions to newly
        # created configuration databases.
        config_file = self.makeDir().joinpath("config")
        with ProfileStore.open(config_file, self.null_profile):
            perms = FilePath(str(config_file)).getPermissions()
            self.assertEqual("rw-------", perms.shorthand())

    def test_open_permissions_existing_database(self):
        # ProfileStore.open() leaves the file permissions of existing
        # configuration databases.
        config_file = self.makeDir().joinpath("config")
        config_file.touch()
        config_file.chmod(0o644)  # u=rw,go=r
        with ProfileStore.open(config_file, self.null_profile):
            perms = FilePath(str(config_file)).getPermissions()
            self.assertEqual("rw-r--r--", perms.shorthand())

    def test_open_does_one_time_migration(self):
        home = self.makeDir()
        dbpath_old = home.joinpath(".maascli.db")
        dbpath_new = home.joinpath(".maas.db")

        # A profile that will be migrated.
        profile = make_profile()

        # Populate the old database with a profile. We're using the new
        # ProfileStore but that's okay; the schemas are compatible.
        with ProfileStore.open(dbpath_old, self.null_profile) as config_old:
            config_old.save(profile)

        # Immediately as we open the new database, profiles from the old
        # database are migrated.
        with ProfileStore.open(dbpath_new, dbpath_old) as config_new:
            self.assertEqual({profile.name}, set(config_new))
            profile_migrated = config_new.load(profile.name)
            self.assertEqual(profile, profile_migrated)
            # Before closing, delete the migrated profile.
            config_new.delete(profile.name)

        # After reopening the new database we see the migrated profile that we
        # deleted has stayed deleted; it has not been migrated a second time.
        with ProfileStore.open(dbpath_new, self.null_profile) as config_new:
            self.assertRaises(ProfileNotFound, config_new.load, profile.name)

        # It is still present and correct in the old database.
        with ProfileStore.open(dbpath_old, self.null_profile) as config_old:
            self.assertEqual(profile, config_old.load(profile.name))


class TestProfileStoreDefault(TestCase):
    """Tests for `ProfileStore` default profile."""

    def test_getting_and_setting_default_profile(self):
        database = sqlite3.connect(":memory:")
        config = ProfileStore(database)
        self.assertIsNone(config.default)
        profile = make_profile()
        config.default = profile
        self.assertEqual(profile, config.default)
        # A side-effect is that the profile is saved.
        self.assertEqual({profile.name}, set(config))

    def test_default_profile_is_persisted(self):
        database = sqlite3.connect(":memory:")
        config1 = ProfileStore(database)
        config2 = ProfileStore(database)
        profile = make_profile()
        config1.default = profile
        self.assertEqual(profile, config2.default)

    def test_default_profile_switch_profile(self):
        database = sqlite3.connect(":memory:")
        config = ProfileStore(database)
        profile1 = make_profile()
        profile2 = make_profile()
        config.default = profile1
        config.default = profile2
        self.assertEqual(profile2, config.default)

    def test_default_profile_remains_default_after_subsequent_save(self):
        database = sqlite3.connect(":memory:")
        profile = make_profile()
        config = ProfileStore(database)
        config.default = profile
        profile = profile.replace(foo="bar")
        config.save(profile)
        self.assertEqual(profile, config.default)

    def test_default_profile_remains_default_after_failed_save(self):
        database = sqlite3.connect(":memory:")
        profile_good = make_profile()
        config = ProfileStore(database)
        config.default = profile_good
        profile_bad = profile_good.replace(foo=object())
        self.assertRaises(TypeError, config.save, profile_bad)
        self.assertEqual(profile_good, config.default)

    def test_clearing_default_profile(self):
        database = sqlite3.connect(":memory:")
        config = ProfileStore(database)
        profile = make_profile()
        config.default = profile
        del config.default
        self.assertIsNone(config.default)
        # The profile itself is not removed.
        self.assertEqual({profile.name}, set(config))

    def test_getting_default_profile_when_profile_has_been_deleted(self):
        database = sqlite3.connect(":memory:")
        config = ProfileStore(database)
        config.default = make_profile()
        config.delete(config.default.name)
        self.assertIsNone(config.default)
        # The profile itself is gone.
        self.assertEqual(set(), set(config))
